import json
import os
import re
from typing import Optional
from typing import TYPE_CHECKING


# TypedDict was added to typing in python 3.8
try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict

from jsonschema import validate

from ddtrace.constants import _SINGLE_SPAN_SAMPLING_MAX_PER_SEC
from ddtrace.constants import _SINGLE_SPAN_SAMPLING_MAX_PER_SEC_NO_LIMIT
from ddtrace.constants import _SINGLE_SPAN_SAMPLING_MECHANISM
from ddtrace.constants import _SINGLE_SPAN_SAMPLING_RATE
from ddtrace.internal.glob_matching import GlobMatcher
from ddtrace.internal.logger import get_logger

from .rate_limiter import RateLimiter


log = get_logger(__name__)

try:
    from json.decoder import JSONDecodeError
except ImportError:
    # handling python 2.X import error
    JSONDecodeError = ValueError  # type: ignore

if TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Text

    from ddtrace.context import Context
    from ddtrace.span import Span

# Big prime number to make hashing better distributed
KNUTH_FACTOR = 1111111111111111111
MAX_SPAN_ID = 2 ** 64


class SamplingMechanism(object):
    DEFAULT = 0
    AGENT_RATE = 1
    REMOTE_RATE = 2
    TRACE_SAMPLING_RULE = 3
    MANUAL = 4
    APPSEC = 5
    REMOTE_RATE_USER = 6
    REMOTE_RATE_DATADOG = 7
    SPAN_SAMPLING_RULE = 8


SAMPLING_DECISION_TRACE_TAG_KEY = "_dd.p.dm"

# Use regex to validate trace tag value
TRACE_TAG_RE = re.compile(r"^-([0-9])$")


SpanSamplingRules = TypedDict(
    "SpanSamplingRules",
    {
        "name": str,
        "service": str,
        "sample_rate": float,
        "max_per_second": int,
    },
    total=False,
)

SPAN_SAMPLING_JSON_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "anyOf": [
            {"properties": {"service": {"type": "string"}}, "required": ["service"]},
            {"properties": {"name": {"type": "string"}}, "required": ["name"]},
        ],
        "properties": {"max_per_second": {"type": "integer"}, "sample_rate": {"type": "number"}},
    },
}


def _set_trace_tag(
    context,  # type: Context
    sampling_mechanism,  # type: int
):
    # type: (...) -> Optional[Text]

    value = "-%d" % sampling_mechanism

    context._meta[SAMPLING_DECISION_TRACE_TAG_KEY] = value

    return value


def _unset_trace_tag(
    context,  # type: Context
):
    # type: (...) -> Optional[Text]
    if SAMPLING_DECISION_TRACE_TAG_KEY not in context._meta:
        return None

    value = context._meta[SAMPLING_DECISION_TRACE_TAG_KEY]
    del context._meta[SAMPLING_DECISION_TRACE_TAG_KEY]
    return value


def validate_sampling_decision(
    meta,  # type: Dict[str, str]
):
    # type: (...) -> Dict[str, str]
    value = meta.get(SAMPLING_DECISION_TRACE_TAG_KEY)
    if value:
        # Skip propagating invalid sampling mechanism trace tag
        if TRACE_TAG_RE.match(value) is None:
            del meta[SAMPLING_DECISION_TRACE_TAG_KEY]
            meta["_dd.propagation_error"] = "decoding_error"
            log.warning("failed to decode _dd.p.dm: %r", value, exc_info=True)
    return meta


def update_sampling_decision(
    context,  # type: Context
    sampling_mechanism,  # type: int
    sampled,  # type: bool
):
    # type: (...) -> Optional[Text]
    # When sampler keeps trace, we need to set sampling decision trace tag.
    # If sampler rejects trace, we need to remove sampling decision trace tag to avoid unnecessary propagation.
    if sampled:
        return _set_trace_tag(context, sampling_mechanism)
    else:
        return _unset_trace_tag(context)


class SpanSamplingRule:
    """A span sampling rule to evaluate and potentially tag each span upon finish."""

    __slots__ = (
        "_service_matcher",
        "_name_matcher",
        "_sample_rate",
        "_max_per_second",
        "_sampling_id_threshold",
        "_limiter",
        "_matcher",
    )

    def __init__(
        self,
        sample_rate,  # type: float
        max_per_second,  # type: int
        service=None,  # type: Optional[str]
        name=None,  # type: Optional[str]
    ):
        self._sample_rate = sample_rate
        self._sampling_id_threshold = self._sample_rate * MAX_SPAN_ID

        self._max_per_second = max_per_second
        self._limiter = RateLimiter(max_per_second)

        # we need to create matchers for the service and/or name pattern provided
        self._service_matcher = GlobMatcher(service) if service is not None else None
        self._name_matcher = GlobMatcher(name) if name is not None else None

    def sample(self, span):
        # type: (Span) -> bool
        if self._sample(span):
            if self._limiter.is_allowed(span.start_ns):
                self.apply_span_sampling_tags(span)
                return True
        return False

    def _sample(self, span):
        # type: (Span) -> bool
        if self._sample_rate == 1:
            return True
        elif self._sample_rate == 0:
            return False

        return ((span.span_id * KNUTH_FACTOR) % MAX_SPAN_ID) <= self._sampling_id_threshold

    def match(self, span):
        # type: (Span) -> bool
        """Determines if the span's service and name match the configured patterns"""
        name = span.name
        service = span.service
        # If a span lacks a name and service, we can't match on it
        if service is None and name is None:
            return False

        # Default to True, as the rule may not have a name or service rule
        # For whichever rules it does have, it will attempt to match on them
        service_match = True
        name_match = True

        if self._service_matcher:
            if service is None:
                return False
            else:
                service_match = self._service_matcher.match(service)
        if self._name_matcher:
            if name is None:
                return False
            else:
                name_match = self._name_matcher.match(name)
        return service_match and name_match

    def apply_span_sampling_tags(self, span):
        # type: (Span) -> None
        span.set_metric(_SINGLE_SPAN_SAMPLING_MECHANISM, SamplingMechanism.SPAN_SAMPLING_RULE)
        span.set_metric(_SINGLE_SPAN_SAMPLING_RATE, self._sample_rate)
        # Only set this tag if it's not the default -1
        if self._max_per_second != _SINGLE_SPAN_SAMPLING_MAX_PER_SEC_NO_LIMIT:
            span.set_metric(_SINGLE_SPAN_SAMPLING_MAX_PER_SEC, self._max_per_second)


def get_span_sampling_rules():
    # type: () -> List[SpanSamplingRule]
    json_rules = _get_span_sampling_json()
    validate(json_rules, SPAN_SAMPLING_JSON_SCHEMA)
    sampling_rules = []
    for rule in json_rules:
        # If sample_rate not specified default to 100%
        sample_rate = rule.get("sample_rate", 1.0)
        service = rule.get("service")
        name = rule.get("name")
        # If max_per_second not specified default to no limit
        max_per_second = rule.get("max_per_second", _SINGLE_SPAN_SAMPLING_MAX_PER_SEC_NO_LIMIT)
        if service:
            _check_unsupported_pattern(service)
        if name:
            _check_unsupported_pattern(name)

        try:
            sampling_rule = SpanSamplingRule(
                sample_rate=sample_rate, service=service, name=name, max_per_second=max_per_second
            )
        except Exception as e:
            raise ValueError("Error creating single span sampling rule {}: {}".format(json.dumps(rule), e))
        sampling_rules.append(sampling_rule)
    return sampling_rules


def _get_span_sampling_json():
    # type: () -> List[Dict[str, Any]]
    env_json_rules = _get_env_json()
    file_json_rules = _get_file_json()

    if env_json_rules and file_json_rules:
        log.warning(
            (
                "DD_SPAN_SAMPLING_RULES and DD_SPAN_SAMPLING_RULES_FILE detected. "
                "Defaulting to DD_SPAN_SAMPLING_RULES value."
            )
        )
        return env_json_rules
    return env_json_rules or file_json_rules or []


def _get_file_json():
    # type: () -> Optional[List[Dict[str, Any]]]
    file_json_raw = os.getenv("DD_SPAN_SAMPLING_RULES_FILE")
    if file_json_raw:
        with open(file_json_raw) as f:
            return _load_span_sampling_json(f.read())
    return None


def _get_env_json():
    # type: () -> Optional[List[Dict[str, Any]]]
    env_json_raw = os.getenv("DD_SPAN_SAMPLING_RULES")
    if env_json_raw:
        return _load_span_sampling_json(env_json_raw)
    return None


def _load_span_sampling_json(raw_json_rules):
    # type: (str) -> List[Dict[str, Any]]
    try:
        json_rules = json.loads(raw_json_rules)
        if not isinstance(json_rules, list):
            raise TypeError("DD_SPAN_SAMPLING_RULES is not list, got %r" % json_rules)
    except JSONDecodeError:
        raise ValueError("Unable to parse DD_SPAN_SAMPLING_RULES=%r" % raw_json_rules)

    return json_rules


def _check_unsupported_pattern(string):
    # type: (str) -> None
    # We don't support pattern bracket expansion or escape character
    unsupported_chars = {"[", "]", "\\"}
    for char in string:
        if char in unsupported_chars:
            raise ValueError("Unsupported Glob pattern found, character:%r is not supported" % char)


def is_single_span_sampled(span):
    # type: (Span) -> bool
    return span.get_metric(_SINGLE_SPAN_SAMPLING_MECHANISM) == SamplingMechanism.SPAN_SAMPLING_RULE



"""
tests for parsing specs.
"""

from bson.son import SON

from ddtrace.contrib.pymongo.parse import parse_spec


def test_empty():
    cmd = parse_spec(SON([]))
    assert cmd is None


def test_create():
    cmd = parse_spec(SON([("create", "foo")]))
    assert cmd.name == "create"
    assert cmd.coll == "foo"
    assert cmd.tags == {}
    assert cmd.metrics == {}


def test_insert():
    spec = SON(
        [
            ("insert", "bla"),
            ("ordered", True),
            ("documents", ["a", "b"]),
        ]
    )
    cmd = parse_spec(spec)
    assert cmd.name == "insert"
    assert cmd.coll == "bla"
    assert cmd.tags == {"mongodb.ordered": True}
    assert cmd.metrics == {"mongodb.documents": 2}


def test_update():
    spec = SON(
        [
            ("update", u"songs"),
            ("ordered", True),
            (
                "updates",
                [
                    SON(
                        [
                            ("q", {"artist": "Neil"}),
                            ("u", {"$set": {"artist": "Shakey"}}),
                            ("multi", True),
                            ("upsert", False),
                        ]
                    )
                ],
            ),
        ]
    )
    cmd = parse_spec(spec)
    assert cmd.name == "update"
    assert cmd.coll == "songs"
    assert cmd.query == {"artist": "Neil"}



import mongoengine

from .trace import WrappedConnect


# Original connect function
_connect = mongoengine.connect


def patch():
    setattr(mongoengine, "connect", WrappedConnect(_connect))


def unpatch():
    setattr(mongoengine, "connect", _connect)



"""
The PynamoDB integration traces all db calls made with the pynamodb
library through the connection API.

Enabling
~~~~~~~~

The PynamoDB integration is enabled automatically when using
:ref:`ddtrace-run<ddtracerun>` or :func:`patch_all()<ddtrace.patch_all>`.

Or use :func:`patch()<ddtrace.patch>` to manually enable the integration::

    import pynamodb
    from ddtrace import patch, config
    patch(pynamodb=True)

Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.pynamodb["service"]

   The service name reported by default for the PynamoDB instance.

   This option can also be set with the ``DD_PYNAMODB_SERVICE`` environment
   variable.

   Default: ``"pynamodb"``

"""


from ...internal.utils.importlib import require_modules


required_modules = ["pynamodb.connection.base"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch

        __all__ = ["patch"]



import molten

from ddtrace import config
from ddtrace.vendor import wrapt

from .. import trace_utils
from ... import Pin
from ...internal.utils.importlib import func_name


MOLTEN_ROUTE = "molten.route"


def trace_wrapped(resource, wrapped, *args, **kwargs):
    pin = Pin.get_from(molten)
    if not pin or not pin.enabled():
        return wrapped(*args, **kwargs)

    with pin.tracer.trace(func_name(wrapped), service=trace_utils.int_service(pin, config.molten), resource=resource):
        return wrapped(*args, **kwargs)


def trace_func(resource):
    """Trace calls to function using provided resource name"""

    @wrapt.function_wrapper
    def _trace_func(wrapped, instance, args, kwargs):
        pin = Pin.get_from(molten)

        if not pin or not pin.enabled():
            return wrapped(*args, **kwargs)

        with pin.tracer.trace(
            func_name(wrapped), service=trace_utils.int_service(pin, config.molten, pin), resource=resource
        ):
            return wrapped(*args, **kwargs)

    return _trace_func


class WrapperComponent(wrapt.ObjectProxy):
    """Tracing of components"""

    def can_handle_parameter(self, *args, **kwargs):
        func = self.__wrapped__.can_handle_parameter
        cname = func_name(self.__wrapped__)
        resource = "{}.{}".format(cname, func.__name__)
        return trace_wrapped(resource, func, *args, **kwargs)

    # TODO[tahir]: the signature of a wrapped resolve method causes DIError to
    # be thrown since parameter types cannot be determined


class WrapperRenderer(wrapt.ObjectProxy):
    """Tracing of renderers"""

    def render(self, *args, **kwargs):
        func = self.__wrapped__.render
        cname = func_name(self.__wrapped__)
        resource = "{}.{}".format(cname, func.__name__)
        return trace_wrapped(resource, func, *args, **kwargs)


class WrapperMiddleware(wrapt.ObjectProxy):
    """Tracing of callable functional-middleware"""

    def __call__(self, *args, **kwargs):
        func = self.__wrapped__.__call__
        resource = func_name(self.__wrapped__)
        return trace_wrapped(resource, func, *args, **kwargs)


class WrapperRouter(wrapt.ObjectProxy):
    """Tracing of router on the way back from a matched route"""

    def match(self, *args, **kwargs):
        # catch matched route and wrap tracer around its handler and set root span resource
        func = self.__wrapped__.match
        route_and_params = func(*args, **kwargs)

        pin = Pin.get_from(molten)
        if not pin or not pin.enabled():
            return route_and_params

        if route_and_params is not None:
            route, params = route_and_params

            route.handler = trace_func(func_name(route.handler))(route.handler)

            # update root span resource while we know the matched route
            resource = "{} {}".format(
                route.method,
                route.template,
            )
            root_span = pin.tracer.current_root_span()
            root_span.resource = resource

            # if no root route set make sure we record it based on this resolved
            # route
            if root_span and not root_span.get_tag(MOLTEN_ROUTE):
                root_span.set_tag(MOLTEN_ROUTE, route.name)

            return route, params

        return route_and_params



import asyncio
from tempfile import NamedTemporaryFile
import time
from typing import Optional

from fastapi import BackgroundTasks
from fastapi import FastAPI
from fastapi import HTTPException
from fastapi import Header
from fastapi.responses import FileResponse
from fastapi.responses import StreamingResponse
from pydantic import BaseModel


fake_secret_token = "DataDog"

fake_db = {
    "foo": {"id": "foo", "name": "Foo", "description": "This item's description is foo."},
    "bar": {"id": "bar", "name": "Bar", "description": "The bartenders"},
    "testUserID": {"userid": "testUserID", "name": "Test User"},
}


class Item(BaseModel):
    id: str
    name: str
    description: Optional[str] = None


class User(BaseModel):
    userid: int
    name: str


def get_app():
    app = FastAPI()

    @app.get("/")
    async def read_homepage(sleep: str = Header(...)):
        if sleep == "True":
            time.sleep(2)
            return {"Homepage Read": "Sleep"}
        return {"Homepage Read": "Success"}

    @app.get("/items/{item_id}", response_model=Item)
    async def read_item(item_id: str, x_token: str = Header(...)):
        if x_token != fake_secret_token:
            raise HTTPException(status_code=401, detail="Invalid X-Token header")
        if item_id not in fake_db:
            raise HTTPException(status_code=404, detail="Item not found")
        return fake_db[item_id]

    @app.post("/items/", response_model=Item)
    async def create_item(item: Item, x_token: str = Header(...)):
        if x_token != fake_secret_token:
            raise HTTPException(status_code=401, detail="Invalid X-Token header")
        if item.id in fake_db:
            raise HTTPException(status_code=400, detail="Item already exists")
        fake_db[item.id] = item
        return item

    @app.get("/users/{userid:str}")
    async def get_user(userid: str, x_token: str = Header(...)):
        if x_token != fake_secret_token:
            raise HTTPException(status_code=401, detail="Invalid X-Token header")
        if userid not in fake_db:
            raise HTTPException(status_code=404, detail="User not found")
        return fake_db[userid]

    @app.get("/users/{userid:str}/info")
    async def get_user_info(userid: str, x_token: str = Header(...)):
        if x_token != fake_secret_token:
            raise HTTPException(status_code=401, detail="Invalid X-Token header")
        if userid not in fake_db:
            raise HTTPException(status_code=404, detail="User not found")
        return {"User Info": "Here"}

    @app.get("/users/{userid:str}/{attribute:str}")
    async def get_user_attribute(userid: str, attribute: str, x_token: str = Header(...)):
        if x_token != fake_secret_token:
            raise HTTPException(status_code=401, detail="Invalid X-Token header")
        if userid not in fake_db:
            raise HTTPException(status_code=404, detail="User not found")
        return {"User Attribute": fake_db[userid].get(attribute, "Fake Attribute")}

    @app.get("/500")
    async def error():
        """
        An example error. Switch the `debug` setting to see either tracebacks or 500 pages.
        """
        raise RuntimeError("Server error")

    @app.get("/stream")
    async def stream():
        def stream_response():
            yield b"streaming"

        return StreamingResponse(stream_response())

    @app.get("/file")
    async def file():
        with NamedTemporaryFile(delete=False) as fp:
            fp.write(b"Datadog says hello!")
            fp.flush()
            return FileResponse(fp.name)

    async def custom_task():
        await asyncio.sleep(1)

    @app.get("/asynctask")
    async def asynctask(bg_tasks: BackgroundTasks):
        bg_tasks.add_task(custom_task)
        return "task added"

    subapp = FastAPI()

    @subapp.get("/hello/{name}")
    def hello():
        return {"Greeting": "Hello"}

    app.mount("/sub-app", subapp)

    return app



# stdlib
import time
from unittest import skipIf

# 3p
import psycopg2
from psycopg2 import extensions
from psycopg2 import extras

from ddtrace import Pin
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.contrib.psycopg.patch import PSYCOPG2_VERSION
from ddtrace.contrib.psycopg.patch import patch
from ddtrace.contrib.psycopg.patch import unpatch
from tests.contrib.config import POSTGRES_CONFIG
from tests.opentracer.utils import init_tracer
from tests.utils import TracerTestCase
from tests.utils import assert_is_measured
from tests.utils import snapshot


if PSYCOPG2_VERSION >= (2, 7):
    from psycopg2.sql import Identifier
    from psycopg2.sql import Literal
    from psycopg2.sql import SQL

TEST_PORT = POSTGRES_CONFIG["port"]


class PsycopgCore(TracerTestCase):

    # default service
    TEST_SERVICE = "postgres"

    def setUp(self):
        super(PsycopgCore, self).setUp()

        patch()

    def tearDown(self):
        super(PsycopgCore, self).tearDown()

        unpatch()

    def _get_conn(self, service=None):
        conn = psycopg2.connect(**POSTGRES_CONFIG)
        pin = Pin.get_from(conn)
        if pin:
            pin.clone(service=service, tracer=self.tracer).onto(conn)

        return conn

    def test_patch_unpatch(self):
        # Test patch idempotence
        patch()
        patch()

        service = "fo"

        conn = self._get_conn(service=service)
        conn.cursor().execute("""select 'blah'""")
        self.assert_structure(dict(name="postgres.query", service=service))
        self.reset()

        # Test unpatch
        unpatch()

        conn = self._get_conn()
        conn.cursor().execute("""select 'blah'""")
        self.assert_has_no_spans()

        # Test patch again
        patch()

        conn = self._get_conn(service=service)
        conn.cursor().execute("""select 'blah'""")
        self.assert_structure(dict(name="postgres.query", service=service))

    def assert_conn_is_traced(self, db, service):

        # ensure the trace pscyopg client doesn't add non-standard
        # methods
        try:
            db.execute("""select 'foobar'""")
        except AttributeError:
            pass

        # Ensure we can run a query and it's correctly traced
        q = """select 'foobarblah'"""

        start = time.time()
        cursor = db.cursor()
        res = cursor.execute(q)
        self.assertIsNone(res)
        rows = cursor.fetchall()
        end = time.time()

        self.assertEquals(rows, [("foobarblah",)])

        self.assert_structure(
            dict(name="postgres.query", resource=q, service=service, error=0, span_type="sql"),
        )
        root = self.get_root_span()
        self.assertIsNone(root.get_tag("sql.query"))
        assert start <= root.start <= end
        assert root.duration <= end - start
        # confirm analytics disabled by default
        self.reset()

        # run a query with an error and ensure all is well
        q = """select * from some_non_existant_table"""
        cur = db.cursor()
        try:
            cur.execute(q)
        except Exception:
            pass
        else:
            assert 0, "should have an error"

        self.assert_structure(
            dict(
                name="postgres.query",
                resource=q,
                service=service,
                error=1,
                span_type="sql",
                meta={
                    "out.host": "127.0.0.1",
                },
                metrics={
                    "out.port": TEST_PORT,
                },
            ),
        )
        root = self.get_root_span()
        assert_is_measured(root)
        self.assertIsNone(root.get_tag("sql.query"))
        self.reset()

    def test_opentracing_propagation(self):
        # ensure OpenTracing plays well with our integration
        query = """SELECT 'tracing'"""

        db = self._get_conn()
        ot_tracer = init_tracer("psycopg-svc", self.tracer)

        with ot_tracer.start_active_span("db.access"):
            cursor = db.cursor()
            cursor.execute(query)
            rows = cursor.fetchall()

        self.assertEquals(rows, [("tracing",)])

        self.assert_structure(
            dict(name="db.access", service="psycopg-svc"),
            (dict(name="postgres.query", resource=query, service="postgres", error=0, span_type="sql"),),
        )
        assert_is_measured(self.get_spans()[1])
        self.reset()

        with self.override_config("psycopg", dict(trace_fetch_methods=True)):
            db = self._get_conn()
            ot_tracer = init_tracer("psycopg-svc", self.tracer)

            with ot_tracer.start_active_span("db.access"):
                cursor = db.cursor()
                cursor.execute(query)
                rows = cursor.fetchall()

            self.assertEquals(rows, [("tracing",)])

            self.assert_structure(
                dict(name="db.access", service="psycopg-svc"),
                (
                    dict(name="postgres.query", resource=query, service="postgres", error=0, span_type="sql"),
                    dict(name="postgres.query.fetchall", resource=query, service="postgres", error=0, span_type="sql"),
                ),
            )
            assert_is_measured(self.get_spans()[1])

    @skipIf(PSYCOPG2_VERSION < (2, 5), "context manager not available in psycopg2==2.4")
    def test_cursor_ctx_manager(self):
        # ensure cursors work with context managers
        # https://github.com/DataDog/dd-trace-py/issues/228
        conn = self._get_conn()
        t = type(conn.cursor())
        with conn.cursor() as cur:
            assert t == type(cur), "{} != {}".format(t, type(cur))
            cur.execute(query="""select 'blah'""")
            rows = cur.fetchall()
            assert len(rows) == 1, rows
            assert rows[0][0] == "blah"

        assert_is_measured(self.get_root_span())
        self.assert_structure(
            dict(name="postgres.query"),
        )

    def test_disabled_execute(self):
        conn = self._get_conn()
        self.tracer.enabled = False
        # these calls were crashing with a previous version of the code.
        conn.cursor().execute(query="""select 'blah'""")
        conn.cursor().execute("""select 'blah'""")
        self.assert_has_no_spans()

    @skipIf(PSYCOPG2_VERSION < (2, 5), "_json is not available in psycopg2==2.4")
    def test_manual_wrap_extension_types(self):
        conn = self._get_conn()
        # NOTE: this will crash if it doesn't work.
        #   _ext.register_type(_ext.UUID, conn_or_curs)
        #   TypeError: argument 2 must be a connection, cursor or None
        extras.register_uuid(conn_or_curs=conn)

        # NOTE: this will crash if it doesn't work.
        #   _ext.register_default_json(conn)
        #   TypeError: argument 2 must be a connection, cursor or None
        extras.register_default_json(conn)

    def test_manual_wrap_extension_adapt(self):
        conn = self._get_conn()
        # NOTE: this will crash if it doesn't work.
        #   items = _ext.adapt([1, 2, 3])
        #   items.prepare(conn)
        #   TypeError: argument 2 must be a connection, cursor or None
        items = extensions.adapt([1, 2, 3])
        items.prepare(conn)

        # NOTE: this will crash if it doesn't work.
        #   binary = _ext.adapt(b'12345)
        #   binary.prepare(conn)
        #   TypeError: argument 2 must be a connection, cursor or None
        binary = extensions.adapt(b"12345")
        binary.prepare(conn)

    @skipIf(PSYCOPG2_VERSION < (2, 7), "quote_ident not available in psycopg2<2.7")
    def test_manual_wrap_extension_quote_ident(self):
        from ddtrace import patch_all

        patch_all()
        from psycopg2.extensions import quote_ident

        # NOTE: this will crash if it doesn't work.
        #   TypeError: argument 2 must be a connection or a cursor
        conn = psycopg2.connect(**POSTGRES_CONFIG)
        quote_ident("foo", conn)

    def test_connect_factory(self):
        services = ["db", "another"]
        for service in services:
            conn = self._get_conn(service=service)
            self.assert_conn_is_traced(conn, service)

    def test_commit(self):
        conn = self._get_conn()
        conn.commit()

        self.assert_structure(dict(name="postgres.connection.commit", service=self.TEST_SERVICE))

    def test_rollback(self):
        conn = self._get_conn()
        conn.rollback()

        self.assert_structure(dict(name="postgres.connection.rollback", service=self.TEST_SERVICE))

    @skipIf(PSYCOPG2_VERSION < (2, 7), "SQL string composition not available in psycopg2<2.7")
    def test_composed_query(self):
        """Checks whether execution of composed SQL string is traced"""
        query = SQL(" union all ").join(
            [SQL("""select {} as x""").format(Literal("one")), SQL("""select {} as x""").format(Literal("two"))]
        )
        db = self._get_conn()

        with db.cursor() as cur:
            cur.execute(query=query)
            rows = cur.fetchall()
            assert len(rows) == 2, rows
            assert rows[0][0] == "one"
            assert rows[1][0] == "two"

        assert_is_measured(self.get_root_span())
        self.assert_structure(
            dict(name="postgres.query", resource=query.as_string(db)),
        )

    @skipIf(PSYCOPG2_VERSION < (2, 7), "SQL string composition not available in psycopg2<2.7")
    def test_composed_query_identifier(self):
        """Checks whether execution of composed SQL string is traced"""
        db = self._get_conn()
        with db.cursor() as cur:
            # DEV: Use a temp table so it is removed after this session
            cur.execute("CREATE TEMP TABLE test (id serial PRIMARY KEY, name varchar(12) NOT NULL UNIQUE);")
            cur.execute("INSERT INTO test (name) VALUES (%s);", ("test_case",))
            spans = self.get_spans()
            assert len(spans) == 2
            self.reset()

            query = SQL("""select {}, {} from {}""").format(Identifier("id"), Identifier("name"), Identifier("test"))
            cur.execute(query=query)
            rows = cur.fetchall()
            assert rows == [(1, "test_case")]

            assert_is_measured(self.get_root_span())
            self.assert_structure(
                dict(name="postgres.query", resource=query.as_string(db)),
            )

    @snapshot()
    @skipIf(PSYCOPG2_VERSION < (2, 7), "SQL string composition not available in psycopg2<2.7")
    def test_composed_query_encoding(self):
        """Checks whether execution of composed SQL string is traced"""
        import logging

        logger = logging.getLogger()
        logger.level = logging.DEBUG
        query = SQL(" union all ").join([SQL("""select 'one' as x"""), SQL("""select 'two' as x""")])
        conn = psycopg2.connect(**POSTGRES_CONFIG)

        with conn.cursor() as cur:
            cur.execute(query=query)
            rows = cur.fetchall()
            assert len(rows) == 2, rows
            assert rows[0][0] == "one"
            assert rows[1][0] == "two"

    def test_analytics_default(self):
        conn = self._get_conn()
        conn.cursor().execute("""select 'blah'""")

        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertIsNone(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    def test_analytics_with_rate(self):
        with self.override_config("psycopg", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
            conn = self._get_conn()
            conn.cursor().execute("""select 'blah'""")

            spans = self.get_spans()
            self.assertEqual(len(spans), 1)
            span = spans[0]
            self.assertEqual(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 0.5)

    def test_analytics_without_rate(self):
        with self.override_config("psycopg", dict(analytics_enabled=True)):
            conn = self._get_conn()
            conn.cursor().execute("""select 'blah'""")

            spans = self.get_spans()
            self.assertEqual(len(spans), 1)
            span = spans[0]
            self.assertEqual(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 1.0)

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_user_specified_app_service(self):
        """
        When a user specifies a service for the app
            The psycopg integration should not use it.
        """
        # Ensure that the service name was configured
        from ddtrace import config

        assert config.service == "mysvc"

        conn = self._get_conn()
        conn.cursor().execute("""select 'blah'""")

        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        assert spans[0].service != "mysvc"

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_PSYCOPG_SERVICE="mysvc"))
    def test_user_specified_service(self):
        conn = self._get_conn()
        conn.cursor().execute("""select 'blah'""")

        spans = self.get_spans()
        self.assertEqual(len(spans), 1)
        assert spans[0].service == "mysvc"

    @skipIf(PSYCOPG2_VERSION < (2, 5), "Connection context managers not defined in <2.5.")
    def test_contextmanager_connection(self):
        service = "fo"
        with self._get_conn(service=service) as conn:
            conn.cursor().execute("""select 'blah'""")
            self.assert_structure(dict(name="postgres.query", service=service))


@skipIf(PSYCOPG2_VERSION < (2, 7), "quote_ident not available in psycopg2<2.7")
def test_manual_wrap_extension_quote_ident_standalone():
    from ddtrace import patch_all

    patch_all()
    from psycopg2.extensions import quote_ident

    # NOTE: this will crash if it doesn't work.
    #   TypeError: argument 2 must be a connection or a cursor
    conn = psycopg2.connect(**POSTGRES_CONFIG)
    quote_ident("foo", conn)



from types import CoroutineType

import pytest

from ddtrace.internal.wrapping import wrap
from tests.internal.py35.asyncstuff import async_func as asyncfoo


def test_wrap_generator_yield_from():
    channel = []

    def wrapper(f, args, kwargs):
        channel[:] = []
        for _ in f(*args, **kwargs):
            channel.append(_)
            yield _

    def g():
        yield from range(10)

    wrap(g, wrapper)

    assert list(g()) == list(range(10)) == channel


@pytest.mark.asyncio
async def test_wrap_coroutine():
    channel = []

    def wrapper(f, args, kwargs):
        async def _handle_coroutine(c):
            retval = await c
            channel.append(retval)
            return retval

        channel[:] = []
        retval = f(*args, **kwargs)
        if isinstance(retval, CoroutineType):
            return _handle_coroutine(retval)
        else:
            channel.append(retval)
            return retval

    async def c():
        return await asyncfoo()

    wrap(c, wrapper)

    assert await c() == 42

    assert channel == [42]


def test_wrap_args_kwarg():
    def f(*args, path=None):
        return (args, path)

    def wrapper(f, args, kwargs):
        return f(*args, **kwargs)

    wrap(f, wrapper)

    assert f(1, 2) == ((1, 2), None)


def test_wrap_arg_args_kwarg_kwargs():
    def f(posarg, *args, path=None, **kwargs):
        return (posarg, args, path, kwargs)

    def wrapper(f, args, kwargs):
        return f(*args, **kwargs)

    wrap(f, wrapper)

    assert f(1, 2) == (1, (2,), None, {})
    assert f(1, 2, 3, foo="bar") == (1, (2, 3), None, {"foo": "bar"})
    assert f(1, 2, 3, path="bar") == (1, (2, 3), "bar", {})
    assert f(1, 2, 3, 4, path="bar", foo="baz") == (1, (2, 3, 4), "bar", {"foo": "baz"})
    assert f(1, path="bar", foo="baz") == (1, (), "bar", {"foo": "baz"})



from requests_mock import Adapter

from ddtrace import config
from tests.utils import TracerTestCase

from .test_requests import BaseRequestTestCase


class TestRequestsDistributed(BaseRequestTestCase, TracerTestCase):
    def headers_here(self, tracer, request, root_span):
        # Use an additional matcher to query the request headers.
        # This is because the parent_id can only been known within such a callback,
        # as it's defined on the requests span, which is not available when calling register_uri
        headers = request.headers
        assert "x-datadog-trace-id" in headers
        assert "x-datadog-parent-id" in headers
        assert str(root_span.trace_id) == headers["x-datadog-trace-id"]
        req_span = tracer.current_span()
        assert "requests.request" == req_span.name
        assert str(req_span.span_id) == headers["x-datadog-parent-id"]
        return True

    def headers_not_here(self, tracer, request):
        headers = request.headers
        assert "x-datadog-trace-id" not in headers
        assert "x-datadog-parent-id" not in headers
        return True

    def test_propagation_default(self):
        # ensure by default, distributed tracing is enabled
        adapter = Adapter()
        self.session.mount("mock", adapter)

        with self.tracer.trace("root") as root:

            def matcher(request):
                return self.headers_here(self.tracer, request, root)

            adapter.register_uri("GET", "mock://datadog/foo", additional_matcher=matcher, text="bar")
            resp = self.session.get("mock://datadog/foo")
            assert 200 == resp.status_code
            assert "bar" == resp.text

    def test_propagation_true_global(self):
        # distributed tracing can be enabled globally
        adapter = Adapter()
        self.session.mount("mock", adapter)

        with self.override_config("requests", dict(distributed_tracing=True)):
            with self.tracer.trace("root") as root:

                def matcher(request):
                    return self.headers_here(self.tracer, request, root)

                adapter.register_uri("GET", "mock://datadog/foo", additional_matcher=matcher, text="bar")
                resp = self.session.get("mock://datadog/foo")
                assert 200 == resp.status_code
                assert "bar" == resp.text

    def test_propagation_false_global(self):
        # distributed tracing can be disabled globally
        adapter = Adapter()
        self.session.mount("mock", adapter)

        with self.override_config("requests", dict(distributed_tracing=False)):
            with self.tracer.trace("root"):

                def matcher(request):
                    return self.headers_not_here(self.tracer, request)

                adapter.register_uri("GET", "mock://datadog/foo", additional_matcher=matcher, text="bar")
                resp = self.session.get("mock://datadog/foo")
                assert 200 == resp.status_code
                assert "bar" == resp.text

    def test_propagation_true(self):
        # ensure distributed tracing can be enabled manually
        cfg = config.get_from(self.session)
        cfg["distributed_tracing"] = True
        adapter = Adapter()
        self.session.mount("mock", adapter)

        with self.tracer.trace("root") as root:

            def matcher(request):
                return self.headers_here(self.tracer, request, root)

            adapter.register_uri("GET", "mock://datadog/foo", additional_matcher=matcher, text="bar")
            resp = self.session.get("mock://datadog/foo")
            assert 200 == resp.status_code
            assert "bar" == resp.text

        spans = self.pop_spans()
        root, req = spans
        assert "root" == root.name
        assert "requests.request" == req.name
        assert root.trace_id == req.trace_id
        assert root.span_id == req.parent_id

    def test_propagation_false(self):
        # ensure distributed tracing can be disabled manually
        cfg = config.get_from(self.session)
        cfg["distributed_tracing"] = False
        adapter = Adapter()
        self.session.mount("mock", adapter)

        with self.tracer.trace("root"):

            def matcher(request):
                return self.headers_not_here(self.tracer, request)

            adapter.register_uri("GET", "mock://datadog/foo", additional_matcher=matcher, text="bar")
            resp = self.session.get("mock://datadog/foo")
            assert 200 == resp.status_code
            assert "bar" == resp.text



"""
The foo integration instruments the bar and baz features of the
foo library.


Enabling
~~~~~~~~

The foo integration is enabled automatically when using
:ref:`ddtrace-run <ddtracerun>` or :func:`patch_all() <ddtrace.patch_all>`.

Or use :func:`patch() <ddtrace.patch>` to manually enable the integration::

    from ddtrace import patch
    patch(foo=True)


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.foo["service"]

   The service name reported by default for foo instances.

   This option can also be set with the ``DD_FOO_SERVICE`` environment
   variable.

   Default: ``"foo"``


Instance Configuration
~~~~~~~~~~~~~~~~~~~~~~

To configure the foo integration on an per-instance basis use the
``Pin`` API::

    import foo
    from ddtrace import Pin

    myfoo = foo.Foo()
    Pin.override(myfoo, service="myfoo")
"""
from ...internal.utils.importlib import require_modules


required_modules = ["foo"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch
        from .patch import unpatch

        __all__ = ["patch", "unpatch"]



import pytest

from ddtrace.settings import Config
from ddtrace.settings import HttpConfig
from ddtrace.settings import IntegrationConfig
from tests.utils import BaseTestCase
from tests.utils import override_env


class TestConfig(BaseTestCase):
    def test_environment_analytics_enabled(self):
        with self.override_env(dict(DD_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED="False")):
            config = Config()
            self.assertFalse(config.analytics_enabled)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED="False")):
            config = Config()
            self.assertFalse(config.analytics_enabled)

    def test_environment_analytics_overrides(self):
        with self.override_env(dict(DD_ANALYTICS_ENABLED="False", DD_TRACE_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED="False", DD_TRACE_ANALYTICS_ENABLED="False")):
            config = Config()
            self.assertFalse(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED="True", DD_TRACE_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED="True", DD_TRACE_ANALYTICS_ENABLED="False")):
            config = Config()
            self.assertFalse(config.analytics_enabled)

    def test_logs_injection(self):
        with self.override_env(dict(DD_LOGS_INJECTION="True")):
            config = Config()
            self.assertTrue(config.logs_injection)

        with self.override_env(dict(DD_LOGS_INJECTION="false")):
            config = Config()
            self.assertFalse(config.logs_injection)

    def test_service(self):
        # If none is provided the default should be ``None``
        with self.override_env(dict()):
            config = Config()
            self.assertEqual(config.service, None)

        with self.override_env(dict(DD_SERVICE="my-service")):
            config = Config()
            self.assertEqual(config.service, "my-service")

    def test_http_config(self):
        config = Config()
        config._add("django", dict())
        assert config.django.trace_query_string is None
        config.http.trace_query_string = True
        assert config.http.trace_query_string is True
        assert config.django.trace_query_string is True

        # Integration usage
        config = Config()
        config._add("django", dict())
        config.django.http.trace_query_string = True
        assert config.http.trace_query_string is None
        assert config.django.trace_query_string is True
        assert config.django.http.trace_query_string is True


class TestHttpConfig(BaseTestCase):
    def test_trace_headers(self):
        http_config = HttpConfig()
        http_config.trace_headers("some_header")
        assert http_config.header_is_traced("some_header")
        assert not http_config.header_is_traced("some_other_header")

    def test_trace_headers_whitelist_case_insensitive(self):
        http_config = HttpConfig()
        http_config.trace_headers("some_header")
        assert http_config.header_is_traced("sOmE_hEaDeR")
        assert not http_config.header_is_traced("some_other_header")

    def test_trace_multiple_headers(self):
        http_config = HttpConfig()
        http_config.trace_headers(["some_header_1", "some_header_2"])
        assert http_config.header_is_traced("some_header_1")
        assert http_config.header_is_traced("some_header_2")
        assert not http_config.header_is_traced("some_header_3")

    def test_empty_entry_do_not_raise_exception(self):
        http_config = HttpConfig()
        http_config.trace_headers("")

        assert not http_config.header_is_traced("some_header_1")

    def test_none_entry_do_not_raise_exception(self):
        http_config = HttpConfig()
        http_config.trace_headers(None)
        assert not http_config.header_is_traced("some_header_1")

    def test_is_header_tracing_configured(self):
        http_config = HttpConfig()
        assert not http_config.is_header_tracing_configured
        http_config.trace_headers("some_header")
        assert http_config.is_header_tracing_configured

    def test_header_is_traced_case_insensitive(self):
        http_config = HttpConfig()
        http_config.trace_headers("sOmE_hEaDeR")
        assert http_config.header_is_traced("SoMe_HeAdEr")
        assert not http_config.header_is_traced("some_other_header")

    def test_header_is_traced_false_for_empty_header(self):
        http_config = HttpConfig()
        http_config.trace_headers("some_header")
        assert not http_config.header_is_traced("")

    def test_header_is_traced_false_for_none_header(self):
        http_config = HttpConfig()
        http_config.trace_headers("some_header")
        assert not http_config.header_is_traced(None)


class TestIntegrationConfig(BaseTestCase):
    def setUp(self):
        self.config = Config()
        self.integration_config = IntegrationConfig(self.config, "test")

    def test_is_a_dict(self):
        assert isinstance(self.integration_config, dict)

    def test_allow_item_access(self):
        self.integration_config["setting"] = "value"

        # Can be accessed both as item and attr accessor
        assert self.integration_config.setting == "value"
        assert self.integration_config["setting"] == "value"

    def test_allow_attr_access(self):
        self.integration_config.setting = "value"

        # Can be accessed both as item and attr accessor
        assert self.integration_config.setting == "value"
        assert self.integration_config["setting"] == "value"

    def test_allow_both_access(self):
        self.integration_config.setting = "value"
        assert self.integration_config["setting"] == "value"
        assert self.integration_config.setting == "value"

        self.integration_config["setting"] = "new-value"
        assert self.integration_config.setting == "new-value"
        assert self.integration_config["setting"] == "new-value"

    def test_allow_configuring_http(self):
        self.integration_config.http.trace_headers("integration_header")
        assert self.integration_config.http.header_is_traced("integration_header")
        assert not self.integration_config.http.header_is_traced("other_header")

    def test_allow_exist_both_global_and_integration_config(self):
        self.config.trace_headers("global_header")
        assert self.integration_config.header_is_traced("global_header")

        self.integration_config.http.trace_headers("integration_header")
        assert self.integration_config.header_is_traced("integration_header")

        assert not self.integration_config.http.header_is_traced("global_header")
        assert not self.config.header_is_traced("integration_header")

    def test_environment_analytics_enabled(self):
        # default
        self.assertFalse(self.config.analytics_enabled)
        self.assertIsNone(self.config.foo.analytics_enabled)

        with self.override_env(dict(DD_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.analytics_enabled)
            self.assertIsNone(config.foo.analytics_enabled)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.analytics_enabled)
            self.assertIsNone(config.foo.analytics_enabled)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 1.0)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED="True")):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 1.0)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED="False")):
            config = Config()
            self.assertFalse(config.foo.analytics_enabled)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED="False")):
            config = Config()
            self.assertFalse(config.foo.analytics_enabled)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED="True", DD_FOO_ANALYTICS_SAMPLE_RATE="0.5")):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 0.5)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED="True", DD_TRACE_FOO_ANALYTICS_SAMPLE_RATE="0.5")):
            config = Config()
            self.assertTrue(config.foo.analytics_enabled)
            self.assertEqual(config.foo.analytics_sample_rate, 0.5)

    def test_analytics_enabled_attribute(self):
        """Confirm environment variable and kwargs are handled properly"""
        ic = IntegrationConfig(self.config, "foo", analytics_enabled=True)
        self.assertTrue(ic.analytics_enabled)

        ic = IntegrationConfig(self.config, "foo", analytics_enabled=False)
        self.assertFalse(ic.analytics_enabled)

        with self.override_env(dict(DD_FOO_ANALYTICS_ENABLED="True")):
            ic = IntegrationConfig(self.config, "foo", analytics_enabled=False)
            self.assertFalse(ic.analytics_enabled)

        with self.override_env(dict(DD_TRACE_FOO_ANALYTICS_ENABLED="True")):
            ic = IntegrationConfig(self.config, "foo", analytics_enabled=False)
            self.assertFalse(ic.analytics_enabled)

    def test_get_analytics_sample_rate(self):
        """Check method for accessing sample rate based on configuration"""
        ic = IntegrationConfig(self.config, "foo", analytics_enabled=True, analytics_sample_rate=0.5)
        self.assertEqual(ic.get_analytics_sample_rate(), 0.5)

        ic = IntegrationConfig(self.config, "foo", analytics_enabled=True)
        self.assertEqual(ic.get_analytics_sample_rate(), 1.0)

        ic = IntegrationConfig(self.config, "foo", analytics_enabled=False)
        self.assertIsNone(ic.get_analytics_sample_rate())

        with self.override_env(dict(DD_ANALYTICS_ENABLED="True")):
            config = Config()
            ic = IntegrationConfig(config, "foo")
            self.assertEqual(ic.get_analytics_sample_rate(use_global_config=True), 1.0)

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED="True")):
            config = Config()
            ic = IntegrationConfig(config, "foo")
            self.assertEqual(ic.get_analytics_sample_rate(use_global_config=True), 1.0)

        with self.override_env(dict(DD_ANALYTICS_ENABLED="False")):
            config = Config()
            ic = IntegrationConfig(config, "foo")
            self.assertIsNone(ic.get_analytics_sample_rate(use_global_config=True))

        with self.override_env(dict(DD_TRACE_ANALYTICS_ENABLED="False")):
            config = Config()
            ic = IntegrationConfig(config, "foo")
            self.assertIsNone(ic.get_analytics_sample_rate(use_global_config=True))

    def test_service(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service is None

    @BaseTestCase.run_in_subprocess(env_overrides=dict(DD_FOO_SERVICE="foo-svc"))
    def test_service_env_var(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service == "foo-svc"

    @BaseTestCase.run_in_subprocess(env_overrides=dict(DD_FOO_SERVICE_NAME="foo-svc"))
    def test_service_name_env_var(self):
        ic = IntegrationConfig(self.config, "foo")
        assert ic.service == "foo-svc"


@pytest.mark.parametrize(
    "global_headers,int_headers,expected",
    (
        (None, None, (False, False, False)),
        ([], None, (False, False, False)),
        (["Header"], None, (True, False, True)),
        (None, ["Header"], (False, True, True)),
        (None, [], (False, False, False)),
        (["Header"], ["Header"], (True, True, True)),
        ([], [], (False, False, False)),
    ),
)
def test_config_is_header_tracing_configured(global_headers, int_headers, expected):
    config = Config()
    integration_config = config.myint

    if global_headers is not None:
        config.trace_headers(global_headers)
    if int_headers is not None:
        integration_config.http.trace_headers(int_headers)

    assert (
        config.http.is_header_tracing_configured,
        integration_config.http.is_header_tracing_configured,
        integration_config.is_header_tracing_configured,
    ) == expected


def test_environment_header_tags():
    with override_env(dict(DD_TRACE_HEADER_TAGS="Host:http.host,User-agent:http.user_agent")):
        config = Config()

    assert config.http.is_header_tracing_configured
    assert config._header_tag_name("Host") == "http.host"
    assert config._header_tag_name("User-agent") == "http.user_agent"
    # Case insensitive
    assert config._header_tag_name("User-Agent") == "http.user_agent"


@pytest.mark.parametrize(
    "env,expected",
    (
        (dict(), (512, True)),
        (dict(DD_TRACE_X_DATADOG_TAGS_MAX_LENGTH="0"), (0, False)),
        (dict(DD_TRACE_X_DATADOG_TAGS_MAX_LENGTH="513"), (ValueError, "Invalid value 513")),
        (dict(DD_TRACE_X_DATADOG_TAGS_MAX_LENGTH="-1"), (ValueError, "Invalid value -1")),
    ),
)
def test_x_datadog_tags(env, expected):
    with override_env(env):
        if expected[0] == ValueError:
            with pytest.raises(expected[0]) as exc:
                _ = Config()
            assert expected[1] in exc.value.args[0]
        else:
            _ = Config()
            assert expected == (_._x_datadog_tags_max_length, _._x_datadog_tags_enabled)



import os
import sys

from flask import Flask
from flask import request

from ddtrace import tracer
from ddtrace.contrib.trace_utils import set_user
from tests.webclient import PingFilter


tracer.configure(
    settings={
        "FILTERS": [PingFilter()],
    }
)
cur_dir = os.path.dirname(os.path.realpath(__file__))
tmpl_path = os.path.join(cur_dir, "test_templates")
app = Flask(__name__, template_folder=tmpl_path)


@app.route("/")
def index():
    return "hello"


@app.route("/identify")
def identify():
    set_user(
        tracer,
        user_id="usr.id",
        email="usr.email",
        name="usr.name",
        session_id="usr.session_id",
        role="usr.role",
        scope="usr.scope",
    )
    return "identify"


@app.route("/shutdown")
def shutdown():
    tracer.shutdown()
    sys.exit(0)


@app.route("/stream")
def hello():
    def resp():
        for i in range(10):
            yield str(i)

    return app.response_class(resp())


@app.route("/body")
def body():
    data = request.get_json()
    return data, 200



from typing import Any
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import Union

from ddtrace.internal.compat import BUILTIN
from ddtrace.internal.compat import PY3
from ddtrace.internal.utils.attrdict import AttrDict
from ddtrace.internal.utils.cache import cached
from ddtrace.vendor import wrapt


NoneType = type(None)

if PY3:
    long = int


def _maybe_slots(obj):
    # type: (Any) -> Union[Tuple[str], List[str]]
    try:
        slots = object.__getattribute__(obj, "__slots__")
        if isinstance(slots, str):
            return (slots,)
        return slots
    except AttributeError:
        return []


@cached()
def _slots(_type):
    # type: (Type) -> Set[str]
    return {_ for cls in object.__getattribute__(_type, "__mro__") for _ in _maybe_slots(cls)}


def get_slots(obj):
    # type: (Any) -> Set[str]
    """Get the object's slots."""
    return _slots(type(obj))


def _isinstance(obj, types):
    # type: (Any, Union[Type, Tuple[Union[Type, Tuple[Any, ...]], ...]]) -> bool
    # DEV: isinstance falls back to calling __getattribute__ which could cause
    # side effects.
    return issubclass(type(obj), types)


class SafeObjectProxy(wrapt.ObjectProxy):
    """Object proxy to make sure we don't call unsafe code.

    Access to the wrapped object is denied to prevent any potential
    side-effects. Arbitrary objects are essentially converted into attribute
    dictionaries. Callable objects are made uncallable to prevent accidental
    calls that can also trigger side-effects.
    """

    def __call__(self, *args, **kwargs):
        # type: (Any, Any) -> Optional[Any]
        raise RuntimeError("Cannot call safe object")

    def __getattribute__(self, name):
        # type: (str) -> Any
        if name == "__wrapped__":
            raise AttributeError("Access denied")

        return super(SafeObjectProxy, self).__getattribute__(name)

    def __getattr__(self, name):
        # type: (str) -> Any
        return type(self).safe(super(SafeObjectProxy, self).__getattr__(name))

    def __getitem__(self, item):
        # type: (Any) -> Any
        return type(self).safe(super(SafeObjectProxy, self).__getitem__(item))

    def __iter__(self):
        # type: () -> Any
        return iter(type(self).safe(_) for _ in super(SafeObjectProxy, self).__iter__())

    def items(self):
        # type: () -> Iterator[Tuple[Any, Any]]
        return (
            (type(self).safe(k), type(self).safe(v)) for k, v in super(SafeObjectProxy, self).__getattr__("items")()
        )

    # Custom object representations might cause side-effects
    def __str__(self):
        return object.__repr__(self)

    __repr__ = __str__

    @classmethod
    def safe(cls, obj):
        # type: (Any) -> Optional[Any]
        """Turn an object into a safe proxy."""
        _type = type(obj)

        if _isinstance(obj, type):
            try:
                if obj.__module__ == BUILTIN:
                    # We are assuming that builtin types are safe
                    return obj
            except AttributeError:
                # No __module__ attribute. We'll use caution
                pass

        elif _type in {str, int, float, bool, NoneType, bytes, complex, long}:
            # We are assuming that scalar builtin type instances are safe
            return obj

        try:
            return cls(AttrDict(object.__getattribute__(obj, "__dict__")))
        except AttributeError:
            pass

        slots = get_slots(obj)
        if slots:
            # Handle slots objects
            return cls(AttrDict({k: object.__getattribute__(obj, k) for k in slots}))

        # raise TypeError("Unhandled object type: %s", type(obj))
        return cls(obj)



from ddtrace.internal.metrics import Metrics


# Debugger metrics
metrics = Metrics(namespace="debugger")

# Metric probe metrics (always enabled)
probe_metrics = Metrics(namespace="debugger.metric")
probe_metrics.enable()



import os
import sys

import aiohttp
import pytest

from ddtrace import Pin
from ddtrace.contrib.aiohttp import patch
from ddtrace.contrib.aiohttp import unpatch
from tests.utils import override_http_config

from ..config import HTTPBIN_CONFIG


HOST = HTTPBIN_CONFIG["host"]
PORT = HTTPBIN_CONFIG["port"]
SOCKET = "{}:{}".format(HOST, PORT)
URL = "http://{}".format(SOCKET)
URL_200 = "{}/status/200".format(URL)
URL_500 = "{}/status/500".format(URL)


@pytest.fixture(autouse=True)
def patch_aiohttp_client():
    patch()
    yield
    unpatch()


@pytest.mark.asyncio
async def test_200_request(snapshot_context):
    with snapshot_context():
        async with aiohttp.ClientSession() as session:
            async with session.get(URL_200) as resp:
                assert resp.status == 200


@pytest.mark.asyncio
async def test_200_request_post(snapshot_context):
    with snapshot_context():
        async with aiohttp.ClientSession() as session:
            async with session.post(URL_200) as resp:
                assert resp.status == 200


@pytest.mark.asyncio
async def test_500_request(snapshot_context):
    with snapshot_context():
        async with aiohttp.ClientSession() as session:
            async with session.get(URL_500) as resp:
                assert resp.status == 500


@pytest.mark.asyncio
async def test_200_request_distributed_tracing(tracer):
    async with aiohttp.ClientSession() as session:
        async with session.get("%s/headers" % URL) as resp:
            assert resp.status == 200
            headers = await resp.json()
            for h in ("x-datadog-trace-id", "x-datadog-parent-id", "x-datadog-sampling-priority"):
                assert h not in headers


@pytest.mark.asyncio
@pytest.mark.parametrize("variant", ["pin", "global"])
async def test_distributed_tracing_disabled(ddtrace_run_python_code_in_subprocess, variant):
    code = """
import asyncio
import sys
import aiohttp
from ddtrace import Pin
from tests.contrib.aiohttp.test_aiohttp_client import URL

async def test():
    async with aiohttp.ClientSession() as session:
        async with session.get("%s/headers" % URL) as resp:
            assert resp.status == 200
            headers = await resp.json()

        for h in ("x-datadog-trace-id", "x-datadog-parent-id", "x-datadog-sampling-priority"):
            assert h not in headers

if sys.version_info >= (3, 7, 0):
    asyncio.run(test())
else:
    asyncio.get_event_loop().run_until_complete(test())
    """
    env = os.environ.copy()
    if variant == "global":
        env["DD_AIOHTTP_CLIENT_DISTRIBUTED_TRACING"] = "false"
    out, err, status, pid = ddtrace_run_python_code_in_subprocess(code, env=env)
    assert status == 0, err
    assert err == b""


@pytest.mark.snapshot(async_mode=False)
def test_configure_global_service_name_env(ddtrace_run_python_code_in_subprocess):
    """
    When only setting DD_SERVICE
        The value from DD_SERVICE is used
    """
    code = """
import asyncio
import sys
import aiohttp
from tests.contrib.aiohttp.test_aiohttp_client import URL_200
async def test():
    async with aiohttp.ClientSession() as session:
        async with session.get(URL_200) as resp:
            pass

if sys.version_info >= (3, 7, 0):
    asyncio.run(test())
else:
    asyncio.get_event_loop().run_until_complete(test())
    """
    env = os.environ.copy()
    env["DD_SERVICE"] = "global-service-name"
    out, err, status, pid = ddtrace_run_python_code_in_subprocess(code, env=env)
    assert status == 0, err
    assert err == b""


@pytest.mark.snapshot(async_mode=False)
def test_configure_service_name_pin(ddtrace_run_python_code_in_subprocess):
    code = """
import asyncio
import sys
import aiohttp
from ddtrace import Pin
from tests.contrib.aiohttp.test_aiohttp_client import URL_200

async def test():
    async with aiohttp.ClientSession() as session:
        Pin.override(session, service="pin-custom-svc")
        async with session.get(URL_200) as resp:
            pass

if sys.version_info >= (3, 7, 0):
    asyncio.run(test())
else:
    asyncio.get_event_loop().run_until_complete(test())
    """
    out, err, status, pid = ddtrace_run_python_code_in_subprocess(code, env=os.environ.copy())
    assert status == 0, err
    assert err == b""


@pytest.mark.asyncio
@pytest.mark.skipif(sys.version_info < (3, 6, 0), reason="Python versions below 3.6 sort dictionaries differently")
async def test_trace_query_string(snapshot_context):
    """
    When trace_query_string is enabled
        The query string is included as a tag on the span
    """
    with override_http_config("aiohttp_client", {"trace_query_string": True}):
        with snapshot_context():
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    "%s/?k1=v1&k2=v2" % URL_200,
                    params={
                        "k3": "v3",
                        "k4": "v4",
                    },
                ) as resp:
                    assert resp.status == 404


@pytest.mark.asyncio
async def test_trace_parenting(snapshot_context):
    pin = Pin.get_from(aiohttp)
    tracer = pin.tracer
    with snapshot_context():
        with tracer.trace("parent"):
            async with aiohttp.ClientSession() as session:
                async with session.get(URL_200) as resp:
                    assert resp.status == 200


@pytest.mark.asyncio
async def test_trace_multiple(snapshot_context):
    with snapshot_context():
        async with aiohttp.ClientSession() as session:
            async with session.get(URL_200) as resp:
                assert resp.status == 200
            async with session.get(URL_200) as resp:
                assert resp.status == 200
            async with session.get(URL_200) as resp:
                assert resp.status == 200



#!/usr/bin/env python
# coding: utf-8

import io

import msgpack

from ddtrace.internal._encoding import Packer


binarydata = bytes(bytearray(range(256)))


def gen_binary_data(idx):
    return binarydata[: idx % 300]


def test_exceeding_unpacker_read_size():
    dumpf = io.BytesIO()

    packer = Packer()

    NUMBER_OF_STRINGS = 6
    read_size = 16
    # 5 ok for read_size=16, while 6 glibc detected *** python: double free or corruption (fasttop):
    # 20 ok for read_size=256, while 25 segfaults / glibc detected *** python: double free or corruption (!prev)
    # 40 ok for read_size=1024, while 50 introduces errors
    # 7000 ok for read_size=1024*1024, while 8000 leads to glibc detected *** python: double free or corruption (!prev):

    for idx in range(NUMBER_OF_STRINGS):
        data = gen_binary_data(idx)
        dumpf.write(packer.pack(data))

    f = io.BytesIO(dumpf.getvalue())
    dumpf.close()

    unpacker = msgpack.Unpacker(f, raw=True, read_size=read_size, use_list=1)

    read_count = 0
    for idx, o in enumerate(unpacker):
        assert type(o) == bytes
        assert o == gen_binary_data(idx)
        read_count += 1

    assert read_count == NUMBER_OF_STRINGS



from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter
from channels.routing import URLRouter
from django.core.asgi import get_asgi_application
from django.urls import re_path

from ddtrace.contrib.asgi import TraceMiddleware


application = get_asgi_application()


async def simple_asgi_app(scope, receive, send):
    await send({"type": "http.response.start", "status": 200, "headers": [(b"Content-Type", b"text/plain")]})
    await send({"type": "http.response.body", "body": b"Hello World. It's me simple asgi app"})


channels_application = ProtocolTypeRouter(
    {
        "http": AuthMiddlewareStack(
            URLRouter(
                [
                    re_path(r"traced-simple-asgi-app/", TraceMiddleware(simple_asgi_app)),
                    re_path(r"simple-asgi-app/", simple_asgi_app),
                    re_path(r"", application),
                ]
            ),
        )
    }
)



import json

import bm

from ddtrace.propagation import _utils as utils
from ddtrace.propagation import http


class HTTPPropagationExtract(bm.Scenario):
    headers = bm.var(type=str)
    extra_headers = bm.var(type=int)
    wsgi_style = bm.var(type=bool)

    def generate_headers(self):
        headers = json.loads(self.headers)
        if self.wsgi_style:
            headers = {utils.get_wsgi_header(header): value for header, value in headers.items()}

        for i in range(self.extra_headers):
            header = "x-test-header-{}".format(i)
            if self.wsgi_style:
                header = utils.get_wsgi_header(header)
            headers[header] = str(i)

        return headers

    def run(self):
        headers = self.generate_headers()

        def _(loops):
            for _ in range(loops):
                http.HTTPPropagator.extract(headers)

        yield _



# -*- encoding: utf-8 -*-
import sys
import typing


try:
    import asyncio
except ImportError:
    DefaultEventLoopPolicy = object

    def get_event_loop_policy():
        # type: (...) -> None
        return None

    def set_event_loop_policy(loop_policy):
        # type: (...) -> None
        pass

    asyncio_available = False

else:
    asyncio_available = True
    DefaultEventLoopPolicy = asyncio.DefaultEventLoopPolicy  # type: ignore[misc]

    get_event_loop_policy = asyncio.get_event_loop_policy  # type: ignore[assignment]
    set_event_loop_policy = asyncio.set_event_loop_policy  # type: ignore[assignment]

    if hasattr(asyncio, "current_task"):
        current_task = asyncio.current_task
    elif hasattr(asyncio.Task, "current_task"):
        current_task = asyncio.Task.current_task  # type: ignore[attr-defined]
    else:

        def current_task(loop=None):
            return None

    if hasattr(asyncio, "all_tasks"):
        all_tasks = asyncio.all_tasks
    elif hasattr(asyncio.Task, "all_tasks"):
        all_tasks = asyncio.Task.all_tasks  # type: ignore[attr-defined]
    else:

        def all_tasks(loop=None):
            return []

    if hasattr(asyncio.Task, "get_name"):
        # `get_name` is only available in Python ≥ 3.8
        def _task_get_name(task):
            return task.get_name()

    else:

        def _task_get_name(task):
            return "Task-%d" % id(task)


from . import _threading


class DdtraceProfilerEventLoopPolicy(DefaultEventLoopPolicy):
    def __init__(self):
        # type: (...) -> None
        super(DdtraceProfilerEventLoopPolicy, self).__init__()
        self.loop_per_thread = _threading._ThreadLink()  # type: _threading._ThreadLink[asyncio.AbstractEventLoop]

    def _clear_threads(self):
        self.loop_per_thread.clear_threads(set(sys._current_frames().keys()))

    def set_event_loop(self, loop):
        super(DdtraceProfilerEventLoopPolicy, self).set_event_loop(loop)
        self._clear_threads()
        if loop is not None:
            self.loop_per_thread.link_object(loop)

    def _ddtrace_get_loop(self, thread_id):
        # type: (...) -> typing.Optional[asyncio.AbstractEventLoop]
        return self.loop_per_thread.get_object(thread_id)



import os
import subprocess
import sys


def run_uwsgi(cmd):
    def _run(*args):
        env = os.environ.copy()
        if sys.version_info[0] == 2:
            # On Python2, it's impossible to import uwsgidecorators without this hack
            env["PYTHONPATH"] = os.path.join(
                os.environ.get("VIRTUAL_ENV", ""),
                "lib",
                "python%s.%s" % (sys.version_info[0], sys.version_info[1]),
                "site-packages",
            )

        return subprocess.Popen(cmd + list(args), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)

    return _run



from types import FunctionType
from typing import Set

from bytecode import Bytecode


def linenos(f):
    # type: (FunctionType) -> Set[int]
    """Get the line numbers of a function."""
    return {instr.lineno for instr in Bytecode.from_code(f.__code__) if hasattr(instr, "lineno")}



import graphene
import pytest

from ddtrace.contrib.graphql import patch
from ddtrace.contrib.graphql import unpatch
from tests.utils import override_config


class Patron(graphene.ObjectType):
    id = graphene.ID()
    name = graphene.String()
    age = graphene.Int()


class PatronQuery(graphene.ObjectType):
    patron = graphene.Field(Patron)

    def resolve_patron(root, info):
        return Patron(id=1, name="Syrus", age=27)


class FailingQuery(graphene.ObjectType):
    patron = graphene.Field(Patron)

    def resolve_patron(root, info):
        raise Exception("exception was raised in a graphene query")


@pytest.fixture(autouse=True)
def enable_graphql_patching():
    patch()
    yield
    unpatch()


@pytest.fixture
def test_source_str():
    return """
    {
      patron {
        id
        name
        age
      }
    }
"""


@pytest.fixture
def enable_graphql_resolvers():
    with override_config("graphql", dict(resolvers_enabled=True)):
        yield


@pytest.fixture
def test_schema():
    return graphene.Schema(query=PatronQuery)


@pytest.fixture
def failing_schema():
    return graphene.Schema(query=FailingQuery)


@pytest.mark.snapshot
def test_schema_execute(test_schema, test_source_str):
    result = test_schema.execute(test_source_str)
    assert not result.errors
    assert result.data == {"patron": {"id": "1", "name": "Syrus", "age": 27}}


@pytest.mark.asyncio
@pytest.mark.snapshot(token="tests.contrib.graphene.test_graphene.test_schema_execute")
@pytest.mark.skipif(graphene.VERSION < (3, 0, 0), reason="execute_async is only supported in graphene>=3.0")
async def test_schema_execute_async(test_schema, test_source_str):
    result = await test_schema.execute_async(test_source_str)
    assert not result.errors
    assert result.data == {"patron": {"id": "1", "name": "Syrus", "age": 27}}


@pytest.mark.snapshot
def test_schema_execute_with_resolvers(test_schema, test_source_str, enable_graphql_resolvers):
    result = test_schema.execute(test_source_str)
    assert not result.errors
    assert result.data == {"patron": {"id": "1", "name": "Syrus", "age": 27}}


@pytest.mark.asyncio
@pytest.mark.snapshot(token="tests.contrib.graphene.test_graphene.test_schema_execute_with_resolvers")
@pytest.mark.skipif(graphene.VERSION < (3, 0, 0), reason="execute_async is only supported in graphene>=3.0")
async def test_schema_execute_async_with_resolvers(test_schema, test_source_str, enable_graphql_resolvers):
    result = await test_schema.execute_async(test_source_str)
    assert not result.errors
    assert result.data == {"patron": {"id": "1", "name": "Syrus", "age": 27}}


@pytest.mark.snapshot(
    ignores=["meta.error.stack"], variants={"v2": graphene.VERSION < (3,), "": graphene.VERSION >= (3,)}
)
def test_schema_failing_execute(failing_schema, test_source_str, enable_graphql_resolvers):
    result = failing_schema.execute(test_source_str)

    assert len(result.errors) == 1
    assert "exception was raised in a graphene query" in str(result.errors[0])



"""
Dynamic Instrumentation
=======================

Configuration
-------------

When using ``ddtrace-run``, dynamic instrumentation can be enabled by setting
the ``DD_DYNAMIC_INSTRUMENTATION_ENABLED`` variable, or programmatically with::

    from ddtrace.debugging import DynamicInstrumentation

    # Enable dynamic instrumentation
    DynamicInstrumentation.enable()

    ...

    # Disable the debugger
    DynamicInstrumentation.disable()
"""

from ddtrace.debugging._debugger import Debugger as DynamicInstrumentation


__all__ = ["DynamicInstrumentation"]



import os
from threading import Thread
from types import FrameType
from types import ModuleType
import typing as t
from warnings import warn

from ddtrace.context import Context
from ddtrace.debugging._config import config
from ddtrace.debugging._debugger import Debugger
from ddtrace.debugging._debugger import DebuggerModuleWatchdog
from ddtrace.debugging._encoding import SnapshotJsonEncoder
from ddtrace.debugging._function.discovery import FunctionDiscovery
from ddtrace.debugging._probe.model import ConditionalProbe
from ddtrace.debugging._probe.model import Probe
from ddtrace.debugging._probe.remoteconfig import ProbePollerEvent
from ddtrace.debugging._snapshot.collector import SnapshotCollector
from ddtrace.debugging._snapshot.collector import SnapshotContext
from ddtrace.debugging._snapshot.model import Snapshot
from ddtrace.internal.compat import ExcInfoType
from ddtrace.internal.module import origin
from ddtrace.internal.utils.formats import asbool


try:
    COLS, _ = os.get_terminal_size()
except Exception:
    COLS = 80
CWD = os.path.abspath(os.getcwd())
TESTS = os.path.join(CWD, "test")
try:
    _ = os.getenv("VIRTUAL_ENV")
    VENV = os.path.abspath(_) if _ is not None else None
except TypeError:
    warn(
        "No virtual environment detected. Running without a virtual environment active might cause exploration tests "
        "to instrument more than intended."
    )
    VENV = None

ENCODE = asbool(os.getenv("DD_DEBUGGER_EXPL_ENCODE", True))


class ModuleCollector(DebuggerModuleWatchdog):
    def __init__(self, *args, **kwargs):
        super(ModuleCollector, self).__init__(*args, **kwargs)

    def on_collect(self, discovery):
        # type: (FunctionDiscovery) -> None
        raise NotImplementedError()

    def after_import(self, module):
        # type: (ModuleType) -> None
        o = origin(module)
        if o.startswith(CWD) and not o.startswith(TESTS) and (VENV is None or not o.startswith(VENV)):
            # We want to instrument only the modules that belong to the codebase
            # and exclude the modules that belong to the tests and the
            # dependencies installed within the virtual environment.
            self.on_collect(FunctionDiscovery(module))

        return super(ModuleCollector, self).after_import(module)


class NoopDebuggerRC(object):
    def __init__(self, *args, **kwargs):
        pass


class NoopService(object):
    def __init__(self, *args, **kwargs):
        pass

    def stop(self):
        pass

    def start(self):
        pass

    def join(self):
        pass


class NoopProbePoller(NoopService):
    pass


class NoopLogsIntakeUploader(NoopService):
    pass


class NoopProbeStatusLogger(object):
    def __init__(self, *args, **kwargs):
        pass

    def received(self, *args, **kwargs):
        pass

    def installed(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass


class NoopSnapshotJsonEncoder(SnapshotJsonEncoder):
    def encode(self, snapshot):
        # type: (Snapshot) -> bytes
        return b""

    @classmethod
    def capture_context(
        cls,
        arguments,  # type: t.List[t.Tuple[str, t.Any]]
        _locals,  # type: t.List[t.Tuple[str, t.Any]]
        throwable,  # type: ExcInfoType
        level=1,  # type: int
    ):
        # type: (...) -> t.Dict[str, t.Any]
        return {}


class ExplorationSnapshotCollector(SnapshotCollector):
    def __init__(self, *args, **kwargs):
        super(ExplorationSnapshotCollector, self).__init__(*args, **kwargs)
        encoder_class = SnapshotJsonEncoder if ENCODE else NoopSnapshotJsonEncoder
        self._encoder = encoder_class("exploration")
        self._encoder._encoders = {Snapshot: self._encoder}
        self._snapshots = []
        self._probes = []
        self._failed_encoding = []
        self.on_snapshot = None

    def _enqueue(self, snapshot):
        # type: (Snapshot) -> None
        if ENCODE:
            try:
                self._snapshots.append(self._encoder.encode(snapshot))
            except Exception:
                self._failed_encoding.append(snapshot)

        self._probes.append(snapshot.probe)
        if self.on_snapshot:
            self.on_snapshot(snapshot)

    def collect(
        self,
        probe,  # type: ConditionalProbe
        frame,  # type: FrameType
        thread,  # type: Thread
        args,  # type: t.List[t.Tuple[str, t.Any]]
        context=None,  # type: t.Optional[Context]
    ):
        # type: (...) -> SnapshotContext
        return SnapshotContext(self, probe, frame, thread, args, context)

    @property
    def snapshots(self):
        # type: () -> t.List[Snapshot]
        return self._snapshots or [None]

    @property
    def probes(self):
        # type: () -> t.List[Probe]
        return self._probes or [None]


class ExplorationDebugger(Debugger):
    __rc__ = NoopDebuggerRC
    __uploader__ = NoopLogsIntakeUploader
    __collector__ = ExplorationSnapshotCollector
    __watchdog__ = ModuleCollector
    __logger__ = NoopProbeStatusLogger
    __poller__ = NoopProbePoller

    @classmethod
    def on_disable(cls):
        # type: () -> None
        raise NotImplementedError()

    @classmethod
    def on_snapshot(cls, snapshot):
        # type: (Snapshot) -> None
        pass

    @classmethod
    def enable(cls):
        # type: () -> None
        config.max_probes = float("inf")
        config.global_rate_limit = float("inf")
        config.metrics = False

        super(ExplorationDebugger, cls).enable()

        cls._instance._collector.on_snapshot = cls.on_snapshot

    @classmethod
    def disable(cls):
        # type: () -> None
        registry = cls._instance._probe_registry

        nprobes = len(registry)
        nokprobes = sum(_.installed for _ in registry.values())

        print(("{:=^%ds}" % COLS).format(" %s: probes stats " % cls.__name__))
        print("")

        print("Installed probes: %d/%d" % (nokprobes, nprobes))
        print("")

        cls.on_disable()

        snapshots = cls.get_snapshots()
        if snapshots and snapshots[-1]:
            print(snapshots[-1].decode())

        super(ExplorationDebugger, cls).disable()

    @classmethod
    def get_snapshots(cls):
        # type: () -> t.List[Snapshot]
        if cls._instance is None:
            return None
        return cls._instance._collector.snapshots

    @classmethod
    def get_triggered_probes(cls):
        # type: () -> t.List[Probe]
        if cls._instance is None:
            return None
        return cls._instance._collector.probes

    @classmethod
    def add_probe(cls, probe):
        # type: (Probe) -> None
        cls._instance._on_configuration(ProbePollerEvent.NEW_PROBES, [probe])

    @classmethod
    def add_probes(cls, probes):
        # type: (t.List[Probe]) -> None
        cls._instance._on_configuration(ProbePollerEvent.NEW_PROBES, probes)

    @classmethod
    def delete_probe(cls, probe):
        # type: (Probe) -> None
        cls._instance._on_configuration(ProbePollerEvent.DELETED_PROBES, [probe])


if asbool(os.getenv("DD_DEBUGGER_EXPL_STATUS_MESSAGES", False)):

    def status(msg):
        # type: (str) -> None
        print(("{:%d}" % COLS).format(msg))


else:

    def status(msg):
        # type: (str) -> None
        pass



import mock
import pytest

from ddtrace.profiling.exporter import _packages
from ddtrace.profiling.exporter._packages import pathlib


@pytest.mark.parametrize(
    "filename,result",
    (
        ("toto.py", True),
        ("blabla/toto.py", True),
        ("/usr/blabla/toto.py", True),
        ("foo.pyc", True),
        ("/usr/foo.pyc", True),
        ("something", False),
        ("/something/", False),
        ("/something/nop", False),
        ("/something/yes.DLL", True),
    ),
)
def test_is_python_source_file(
    filename,  # type: str
    result,  # type: bool
):
    # type: (...) -> None
    assert _packages._is_python_source_file(pathlib.Path(filename)) == result


@mock.patch.object(_packages, "_build_package_file_mapping")
def test_filename_to_package_failure(
    build,  # type: mock.MagicMock
    caplog,  # type: pytest.LogCaptureFixture
):
    # type: (...) -> None
    _packages._FILE_PACKAGE_MAPPING = None

    def _raise():
        raise Exception

    build.side_effect = _raise

    # type: (...) -> None
    assert _packages.filename_to_package(_packages.__file__) is None
    assert len(caplog.records) == 1
    assert caplog.records[0].message == (
        "Unable to build package file mapping, please report this to https://github.com/DataDog/dd-trace-py/issues"
    )


def test_filename_to_package():
    # type: (...) -> None
    _packages._FILE_PACKAGE_MAPPING = None

    package = _packages.filename_to_package(_packages.__file__)
    assert package.name == "ddtrace"
    package = _packages.filename_to_package(pytest.__file__)
    assert package.name == "pytest"

    import six

    package = _packages.filename_to_package(six.__file__)
    assert package.name == "six"

    import google.protobuf.internal as gp

    package = _packages.filename_to_package(gp.__file__)
    assert package.name == "protobuf"



from typing import List
from typing import Tuple

from ...constants import ENV_KEY
from .collector import ValueCollector
from .constants import LANG
from .constants import LANG_INTERPRETER
from .constants import LANG_VERSION
from .constants import SERVICE
from .constants import TRACER_VERSION


class RuntimeTagCollector(ValueCollector):
    periodic = False
    value = []  # type: List[Tuple[str, str]]


class TracerTagCollector(RuntimeTagCollector):
    """Tag collector for the ddtrace Tracer"""

    required_modules = ["ddtrace"]

    def collect_fn(self, keys):
        ddtrace = self.modules.get("ddtrace")
        # make sure to copy _services to avoid RuntimeError: Set changed size during iteration
        tags = [(SERVICE, service) for service in list(ddtrace.tracer._services)]
        if ENV_KEY in ddtrace.tracer._tags:
            tags.append((ENV_KEY, ddtrace.tracer._tags.get(ENV_KEY)))
        return tags


class PlatformTagCollector(RuntimeTagCollector):
    """Tag collector for the Python interpreter implementation.

    Tags collected:
    - ``lang_interpreter``:

      * For CPython this is 'CPython'.
      * For Pypy this is ``PyPy``
      * For Jython this is ``Jython``

    - `lang_version``,  eg ``2.7.10``
    - ``lang`` e.g. ``Python``
    - ``tracer_version`` e.g. ``0.29.0``

    """

    required_modules = ["platform", "ddtrace"]

    def collect_fn(self, keys):
        platform = self.modules.get("platform")
        ddtrace = self.modules.get("ddtrace")
        tags = [
            (LANG, "python"),
            (LANG_INTERPRETER, platform.python_implementation()),
            (LANG_VERSION, platform.python_version()),
            (TRACER_VERSION, ddtrace.__version__),
        ]
        return tags



import mock
import pytest
import urllib3

from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import ERROR_MSG
from ddtrace.constants import ERROR_STACK
from ddtrace.constants import ERROR_TYPE
from ddtrace.contrib.urllib3 import patch
from ddtrace.contrib.urllib3 import unpatch
from ddtrace.ext import http
from ddtrace.pin import Pin
from tests.opentracer.utils import init_tracer
from tests.utils import TracerTestCase
from tests.utils import snapshot


# host:port of httpbin_local container
HOST = "localhost"
PORT = 8001
SOCKET = "{}:{}".format(HOST, PORT)
URL_200 = "http://{}/status/200".format(SOCKET)
URL_500 = "http://{}/status/500".format(SOCKET)


class BaseUrllib3TestCase(TracerTestCase):
    """Provides the setup and teardown for patching/unpatching the urllib3 integration"""

    def setUp(self):
        super(BaseUrllib3TestCase, self).setUp()

        patch()
        self.http = urllib3.PoolManager()
        Pin.override(urllib3.connectionpool.HTTPConnectionPool, tracer=self.tracer)

    def tearDown(self):
        super(BaseUrllib3TestCase, self).tearDown()
        unpatch()


class TestUrllib3(BaseUrllib3TestCase):
    def test_HTTPConnectionPool_traced(self):
        """Tests that requests made from the HTTPConnectionPool are traced"""
        pool = urllib3.connectionpool.HTTPConnectionPool(HOST, PORT)
        # Test a relative URL
        r = pool.request("GET", "/status/200")
        assert r.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag(http.URL) == URL_200

        # Test an absolute URL
        r = pool.request("GET", URL_200)
        assert r.status == 200
        assert len(self.pop_spans()) == 1

    def test_traced_connection_from_url(self):
        """Tests tracing from ``connection_from_url`` is set up"""
        conn = urllib3.connectionpool.connection_from_url(URL_200)
        resp = conn.request("GET", "/")
        assert resp.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag(http.URL) == "http://" + SOCKET + "/"

    def test_resource_path(self):
        """Tests that a successful request tags a single span with the URL"""
        resp = self.http.request("GET", URL_200)
        assert resp.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag("http.url") == URL_200

    def test_tracer_disabled(self):
        """Tests a disabled tracer produces no spans on request"""
        self.tracer.enabled = False
        out = self.http.request("GET", URL_200)
        assert out.status == 200
        spans = self.pop_spans()
        assert len(spans) == 0

    def test_args_kwargs(self):
        """
        Test that args are kwargs are correctly inferred from the target function's
        signature.

        The args/kwargs used in the integration are:
            - method (idx 0)
            - url (idx 1)
            - headers (idx 3)
        """
        inputs = [
            (("POST", URL_200, b"payload", {"accept": "*"}), {}),
            (("POST", URL_200, b"payload"), {"headers": {"accept": "*"}}),
            (("POST", URL_200), {"headers": {"accept": "*"}}),
            (("POST",), {"url": URL_200, "headers": {"accept": "*"}}),
            ((), {"method": "POST", "url": URL_200, "headers": {"accept": "*"}}),
        ]

        for args, kwargs in inputs:

            with self.override_http_config("urllib3", {"_header_tags": dict()}):
                config.urllib3.http.trace_headers(["accept"])
                pool = urllib3.connectionpool.HTTPConnectionPool(HOST, PORT)
                out = pool.urlopen(*args, **kwargs)
            assert out.status == 200
            spans = self.pop_spans()
            assert len(spans) == 1
            s = spans[0]
            assert s.get_tag(http.METHOD) == "POST"
            assert s.get_tag(http.STATUS_CODE) == "200"
            assert s.get_tag(http.URL) == URL_200
            assert s.get_tag("http.request.headers.accept") == "*"

    def test_untraced_request(self):
        """Disabling tracing with unpatch should submit no spans"""
        # Assumes patching is done in the setUp of the test
        unpatch()

        out = self.http.request("GET", URL_200)
        assert out.status == 200
        spans = self.pop_spans()
        assert len(spans) == 0

    def test_double_patch(self):
        """Ensure that double patch doesn't duplicate instrumentation"""
        patch()
        connpool = urllib3.connectionpool.HTTPConnectionPool(HOST, PORT)
        setattr(connpool, "datadog_tracer", self.tracer)

        out = connpool.urlopen("GET", URL_200)
        assert out.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1

    def test_200(self):
        """Test 200 span tags"""
        out = self.http.request("GET", URL_200)
        assert out.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag(http.METHOD) == "GET"
        assert s.get_tag(http.URL) == URL_200
        assert s.get_tag(http.STATUS_CODE) == "200"
        assert s.error == 0
        assert s.span_type == "http"
        assert http.QUERY_STRING not in s.get_tags()

    def test_200_query_string(self):
        """Tests query string tag is added when trace_query_string config is set"""
        query_string = "key=value&key2=value2"
        URL_200_QS = URL_200 + "?" + query_string
        with self.override_http_config("urllib3", dict(trace_query_string=True)):
            out = self.http.request("GET", URL_200_QS)
            assert out.status == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag(http.METHOD) == "GET"
        assert s.get_tag(http.STATUS_CODE) == "200"
        assert s.get_tag(http.URL) == URL_200_QS
        assert s.error == 0
        assert s.span_type == "http"
        assert s.get_tag(http.QUERY_STRING) == query_string

    def test_post_500(self):
        """Test a request with method POST and expected status 500"""
        out = self.http.request("POST", URL_500)
        assert out.status == 500
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag(http.METHOD) == "POST"
        assert s.get_tag(http.STATUS_CODE) == "500"
        assert s.get_tag(http.URL) == URL_500
        assert s.error == 1

    def test_connection_retries(self):
        """Tests a connection error results in error spans with proper exc info"""
        retries = 3
        try:
            self.http.request("GET", "http://localhost:9999", retries=retries)
        except Exception:
            pass
        else:
            assert 0, "expected error"

        spans = self.pop_spans()
        assert len(spans) == 4  # Default retry behavior is 3 retries + original request
        for i, s in enumerate(spans):
            assert s.get_tag(http.METHOD) == "GET"
            if i > 0:
                assert s.get_tag(http.RETRIES_REMAIN) == str(retries - i)
            assert s.error == 1
            assert "Failed to establish a new connection" in s.get_tag(ERROR_MSG)
            assert "Failed to establish a new connection" in s.get_tag(ERROR_STACK)
            assert "Traceback (most recent call last)" in s.get_tag(ERROR_STACK)
            assert "urllib3.exceptions.MaxRetryError" in s.get_tag(ERROR_TYPE)

    def test_default_service_name(self):
        """Test the default service name is set"""
        out = self.http.request("GET", URL_200)
        assert out.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.service == "urllib3"

    def test_user_set_service_name(self):
        """Test the user-set service name is set on the span"""
        with self.override_config("urllib3", dict(split_by_domain=False, service_name="clients")):
            out = self.http.request("GET", URL_200)
        assert out.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "clients"

    def test_parent_service_name_split_by_domain(self):
        """
        Tests the request span does not inherit the service name when
        split_by_domain is set to True
        """
        with self.override_config("urllib3", dict(split_by_domain=True)):
            with self.tracer.trace("parent.span", service="web"):
                out = self.http.request("GET", URL_200)
                assert out.status == 200

        spans = self.pop_spans()
        assert len(spans) == 2
        s = spans[1]

        assert s.name == "urllib3.request"
        assert s.service == SOCKET

    def test_parent_without_service_name(self):
        """Test that span with a parent with no service defaults to the hostname"""
        with self.override_config("urllib3", dict(split_by_domain=True)):
            with self.tracer.trace("parent.span"):
                out = self.http.request("GET", URL_200)
                assert out.status == 200

            spans = self.pop_spans()
            assert len(spans) == 2
            s = spans[1]

            assert s.name == "urllib3.request"
            assert s.service == SOCKET

    def test_split_by_domain_remove_auth_in_url(self):
        """Tests that only the hostname is used as the default service name"""
        with self.override_config("urllib3", dict(split_by_domain=True)):
            out = self.http.request("GET", "http://user:pass@{}".format(SOCKET))
            assert out.status == 200

            spans = self.pop_spans()
            assert len(spans) == 1
            s = spans[0]

            assert s.service == SOCKET

    def test_split_by_domain_includes_port(self):
        """Test the port is included if not 80 or 443"""
        with self.override_config("urllib3", dict(split_by_domain=True)):
            with pytest.raises(Exception):
                # Using a port the service is not listening on will throw an error, which is fine
                self.http.request("GET", "http://httpbin.org:8000/hello", timeout=0.0001, retries=0)

            spans = self.pop_spans()
            assert len(spans) == 1
            s = spans[0]

            assert s.error == 1
            assert s.service == "httpbin.org:8000"

    def test_200_ot(self):
        """OpenTracing version of test_200."""

        ot_tracer = init_tracer("urllib3_svc", self.tracer)

        with ot_tracer.start_active_span("urllib3_get"):
            out = self.http.request("GET", URL_200)
            assert out.status == 200

        spans = self.pop_spans()
        assert len(spans) == 2

        ot_span, dd_span = spans

        # confirm the parenting
        assert ot_span.parent_id is None
        assert dd_span.parent_id == ot_span.span_id

        assert ot_span.name == "urllib3_get"
        assert ot_span.service == "urllib3_svc"

        assert dd_span.get_tag(http.METHOD) == "GET"
        assert dd_span.get_tag(http.STATUS_CODE) == "200"
        assert dd_span.error == 0
        assert dd_span.span_type == "http"

    def test_request_and_response_headers(self):
        """Tests the headers are added as tag when the headers are whitelisted"""
        self.http.request("GET", URL_200, headers={"my-header": "my_value"})
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag("http.request.headers.my-header") is None
        assert s.get_tag("http.response.headers.access-control-allow-origin") is None

        # Enabled when explicitly configured
        with self.override_http_config("urllib3", {"_header_tags": dict()}):
            config.urllib3.http.trace_headers(["my-header", "access-control-allow-origin"])
            self.http.request("GET", URL_200, headers={"my-header": "my_value"})
            spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag("http.request.headers.my-header") == "my_value"
        assert s.get_tag("http.response.headers.access-control-allow-origin") == "*"

    def test_analytics_integration_default(self):
        """Tests the default behavior of analytics integration is disabled"""
        r = self.http.request("GET", URL_200)
        assert r.status == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_metric(ANALYTICS_SAMPLE_RATE_KEY) is None

    def test_analytics_integration_disabled(self):
        """Test disabling the analytics integration"""
        with self.override_config("urllib3", dict(analytics_enabled=False, analytics_sample_rate=0.5)):
            self.http.request("GET", URL_200)

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_metric(ANALYTICS_SAMPLE_RATE_KEY) is None

    def test_analytics_integration_enabled(self):
        """Tests enabline the analytics integration"""
        with self.override_config("urllib3", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
            self.http.request("GET", URL_200)

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_metric(ANALYTICS_SAMPLE_RATE_KEY) == 0.5

    def test_distributed_tracing_enabled(self):
        """Tests distributed tracing headers are passed by default"""
        # Check that distributed tracing headers are passed down; raise an error rather than make the
        # request since we don't care about the response at all
        config.urllib3["distributed_tracing"] = True
        with mock.patch(
            "urllib3.connectionpool.HTTPConnectionPool._make_request", side_effect=ValueError
        ) as m_make_request:
            with pytest.raises(ValueError):
                self.http.request("GET", URL_200)

            spans = self.pop_spans()
            s = spans[0]
            expected_headers = {
                "x-datadog-trace-id": str(s.trace_id),
                "x-datadog-parent-id": str(s.span_id),
                "x-datadog-sampling-priority": "1",
                "x-datadog-tags": "_dd.p.dm=-0",
            }
            m_make_request.assert_called_with(
                mock.ANY, "GET", "/status/200", body=None, chunked=mock.ANY, headers=expected_headers, timeout=mock.ANY
            )

    def test_distributed_tracing_disabled(self):
        """Test with distributed tracing disabled does not propagate the headers"""
        config.urllib3["distributed_tracing"] = False
        with mock.patch(
            "urllib3.connectionpool.HTTPConnectionPool._make_request", side_effect=ValueError
        ) as m_make_request:
            with pytest.raises(ValueError):
                self.http.request("GET", URL_200)

            m_make_request.assert_called_with(
                mock.ANY, "GET", "/status/200", body=None, chunked=mock.ANY, headers={}, timeout=mock.ANY
            )


@pytest.fixture()
def patch_urllib3():
    patch()
    try:
        yield
    finally:
        unpatch()


@snapshot()
def test_urllib3_poolmanager_snapshot(patch_urllib3):
    pool = urllib3.PoolManager()
    pool.request("GET", URL_200)


@snapshot()
def test_urllib3_connectionpool_snapshot(patch_urllib3):
    pool = urllib3.connectionpool.HTTPConnectionPool(HOST, PORT)
    pool.request("GET", "/status/200")



"""
To trace sqlalchemy queries, add instrumentation to the engine class
using the patch method that **must be called before** importing sqlalchemy::

    # patch before importing `create_engine`
    from ddtrace import Pin, patch
    patch(sqlalchemy=True)

    # use SQLAlchemy as usual
    from sqlalchemy import create_engine

    engine = create_engine('sqlite:///:memory:')
    engine.connect().execute("SELECT COUNT(*) FROM users")

    # Use a PIN to specify metadata related to this engine
    Pin.override(engine, service='replica-db')
"""
from ...internal.utils.importlib import require_modules


required_modules = ["sqlalchemy", "sqlalchemy.event"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .engine import trace_engine
        from .patch import patch
        from .patch import unpatch

        __all__ = ["trace_engine", "patch", "unpatch"]



import aiobotocore.session
from async_generator import async_generator
from async_generator import asynccontextmanager
from async_generator import yield_

from ddtrace import Pin


LOCALSTACK_ENDPOINT_URL = {
    "s3": "http://localhost:5000",
    "ec2": "http://localhost:5001",
    "kms": "http://localhost:5002",
    "sqs": "http://localhost:5003",
    "lambda": "http://localhost:5004",
    "kinesis": "http://localhost:5005",
}


@asynccontextmanager
@async_generator
async def aiobotocore_client(service, tracer):
    """Helper function that creates a new aiobotocore client so that
    it is closed at the end of the context manager.
    """
    session = aiobotocore.session.get_session()
    endpoint = LOCALSTACK_ENDPOINT_URL[service]
    client = session.create_client(
        service,
        region_name="us-west-2",
        endpoint_url=endpoint,
        aws_access_key_id="aws",
        aws_secret_access_key="aws",
        aws_session_token="aws",
    )

    """Check that ClientCreatorContext exists and that client is an expected type before async with
    ClientCreatorContext was added in aiobotocore 1.x: https://github.com/aio-libs/aiobotocore/pull/659
    In 0.x, client evaluates to aiobotocore.client.EC2 while in 1.x, client
    evaluates to aiobotocore.session.ClientCreatorContext
    """
    if hasattr(aiobotocore.session, "ClientCreatorContext") and isinstance(
        client, aiobotocore.session.ClientCreatorContext
    ):
        async with client as client:
            Pin.override(client, tracer=tracer)
            await yield_(client)

    else:
        Pin.override(client, tracer=tracer)
        try:
            await yield_(client)
        finally:
            await client.close()



import sys

import pymemcache
from pymemcache.client.base import Client
from pymemcache.exceptions import MemcacheClientError
from pymemcache.exceptions import MemcacheIllegalInputError
from pymemcache.exceptions import MemcacheServerError
from pymemcache.exceptions import MemcacheUnknownCommandError
from pymemcache.exceptions import MemcacheUnknownError

# 3p
from ddtrace import config
from ddtrace.vendor import wrapt

# project
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import memcached as memcachedx
from ...ext import net
from ...internal.compat import reraise
from ...internal.logger import get_logger
from ...pin import Pin


log = get_logger(__name__)


# keep a reference to the original unpatched clients
_Client = Client


class WrappedClient(wrapt.ObjectProxy):
    """Wrapper providing patched methods of a pymemcache Client.

    Relevant connection information is obtained during initialization and
    attached to each span.

    Keys are tagged in spans for methods that act upon a key.
    """

    def __init__(self, *args, **kwargs):
        c = _Client(*args, **kwargs)
        super(WrappedClient, self).__init__(c)

        # tags to apply to each span generated by this client
        tags = _get_address_tags(*args, **kwargs)

        parent_pin = Pin.get_from(pymemcache)

        if parent_pin:
            pin = parent_pin.clone(tags=tags)
        else:
            pin = Pin(tags=tags)

        # attach the pin onto this instance
        pin.onto(self)

    def set(self, *args, **kwargs):
        return self._traced_cmd("set", *args, **kwargs)

    def set_many(self, *args, **kwargs):
        return self._traced_cmd("set_many", *args, **kwargs)

    def add(self, *args, **kwargs):
        return self._traced_cmd("add", *args, **kwargs)

    def replace(self, *args, **kwargs):
        return self._traced_cmd("replace", *args, **kwargs)

    def append(self, *args, **kwargs):
        return self._traced_cmd("append", *args, **kwargs)

    def prepend(self, *args, **kwargs):
        return self._traced_cmd("prepend", *args, **kwargs)

    def cas(self, *args, **kwargs):
        return self._traced_cmd("cas", *args, **kwargs)

    def get(self, *args, **kwargs):
        return self._traced_cmd("get", *args, **kwargs)

    def get_many(self, *args, **kwargs):
        return self._traced_cmd("get_many", *args, **kwargs)

    def gets(self, *args, **kwargs):
        return self._traced_cmd("gets", *args, **kwargs)

    def gets_many(self, *args, **kwargs):
        return self._traced_cmd("gets_many", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._traced_cmd("delete", *args, **kwargs)

    def delete_many(self, *args, **kwargs):
        return self._traced_cmd("delete_many", *args, **kwargs)

    def incr(self, *args, **kwargs):
        return self._traced_cmd("incr", *args, **kwargs)

    def decr(self, *args, **kwargs):
        return self._traced_cmd("decr", *args, **kwargs)

    def touch(self, *args, **kwargs):
        return self._traced_cmd("touch", *args, **kwargs)

    def stats(self, *args, **kwargs):
        return self._traced_cmd("stats", *args, **kwargs)

    def version(self, *args, **kwargs):
        return self._traced_cmd("version", *args, **kwargs)

    def flush_all(self, *args, **kwargs):
        return self._traced_cmd("flush_all", *args, **kwargs)

    def quit(self, *args, **kwargs):
        return self._traced_cmd("quit", *args, **kwargs)

    def set_multi(self, *args, **kwargs):
        """set_multi is an alias for set_many"""
        return self._traced_cmd("set_many", *args, **kwargs)

    def get_multi(self, *args, **kwargs):
        """set_multi is an alias for set_many"""
        return self._traced_cmd("get_many", *args, **kwargs)

    def _traced_cmd(self, method_name, *args, **kwargs):
        """Run and trace the given command.

        Any pymemcache exception is caught and span error information is
        set. The exception is then reraised for the application to handle
        appropriately.

        Relevant tags are set in the span.
        """
        method = getattr(self.__wrapped__, method_name)
        p = Pin.get_from(self)

        # if the pin does not exist or is not enabled, shortcut
        if not p or not p.enabled():
            return method(*args, **kwargs)

        with p.tracer.trace(
            memcachedx.CMD,
            service=p.service,
            resource=method_name,
            span_type=SpanTypes.CACHE,
        ) as span:
            span.set_tag(SPAN_MEASURED_KEY)
            # set analytics sample rate
            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.pymemcache.get_analytics_sample_rate())

            # try to set relevant tags, catch any exceptions so we don't mess
            # with the application
            try:
                span.set_tags(p.tags)
                vals = _get_query_string(args)
                query = "{}{}{}".format(method_name, " " if vals else "", vals)
                span.set_tag(memcachedx.QUERY, query)
            except Exception:
                log.debug("Error setting relevant pymemcache tags")

            try:
                return method(*args, **kwargs)
            except (
                MemcacheClientError,
                MemcacheServerError,
                MemcacheUnknownCommandError,
                MemcacheUnknownError,
                MemcacheIllegalInputError,
            ):
                (typ, val, tb) = sys.exc_info()
                span.set_exc_info(typ, val, tb)
                reraise(typ, val, tb)


def _get_address_tags(*args, **kwargs):
    """Attempt to get host and port from args passed to Client initializer."""
    tags = {}
    try:
        if len(args):
            host, port = args[0]
            tags[net.TARGET_HOST] = host
            tags[net.TARGET_PORT] = port
    except Exception:
        log.debug("Error collecting client address tags")

    return tags


def _get_query_string(args):
    """Return the query values given the arguments to a pymemcache command.

    If there are multiple query values, they are joined together
    space-separated.
    """
    keys = ""

    # shortcut if no args
    if not args:
        return keys

    # pull out the first arg which will contain any key
    arg = args[0]

    # if we get a dict, convert to list of keys
    if type(arg) is dict:
        arg = list(arg)

    if type(arg) is str:
        keys = arg
    elif type(arg) is bytes:
        keys = arg.decode()
    elif type(arg) is list and len(arg):
        if type(arg[0]) is str:
            keys = " ".join(arg)
        elif type(arg[0]) is bytes:
            keys = b" ".join(arg).decode()

    return keys



#!/usr/bin/env python
# coding: utf-8

from collections import namedtuple

import pytest

from ddtrace.internal._encoding import packb


class MyList(list):
    pass


class MyDict(dict):
    pass


class MyTuple(tuple):
    pass


MyNamedTuple = namedtuple("MyNamedTuple", "x y")


def test_types():
    with pytest.raises(TypeError):
        assert packb(MyDict()) == packb(dict())
    with pytest.raises(TypeError):
        assert packb(MyList()) == packb(list())
    with pytest.raises(TypeError):
        assert packb(MyNamedTuple(1, 2)) == packb((1, 2))



from typing import Dict

class TagsetDecodeError(ValueError): ...
class TagsetEncodeError(ValueError): ...

class TagsetMaxSizeEncodeError(TagsetEncodeError):
    values: Dict[str, str]
    max_size: int
    current_results: str
    def __init__(self, values: Dict[str, str], max_size: int, current_results: str): ...

class TagsetMaxSizeDecodeError(TagsetDecodeError):
    value: Dict[str, str]
    max_size: int
    def __init__(self, value: Dict[str, str], max_size: int): ...

def decode_tagset_string(tagset: str) -> Dict[str, str]: ...
def encode_tagset_values(values: Dict[str, str], max_size: int = 512) -> str: ...



from __future__ import absolute_import

from typing import Callable
from typing import Optional


class uWSGIConfigError(Exception):
    """uWSGI configuration error.

    This is raised when uwsgi configuration is incompatible with the library.
    """


class uWSGIMasterProcess(Exception):
    """The process is uWSGI master process."""


def check_uwsgi(worker_callback=None, atexit=None):
    # type: (Optional[Callable], Optional[Callable]) -> None
    """Check whetever uwsgi is running and what needs to be done.

    :param worker_callback: Callback function to call in uWSGI worker processes.
    """
    try:
        import uwsgi
    except ImportError:
        return

    if not uwsgi.opt.get("enable-threads"):
        raise uWSGIConfigError("enable-threads option must be set to true")

    # If uwsgi has more than one process, it is running in prefork operational mode: uwsgi is going to fork multiple
    # sub-processes.
    # If lazy-app is enabled, then the app is loaded in each subprocess independently. This is fine.
    # If it's not enabled, then the app will be loaded in the master process, and uwsgi will `fork()` abruptly,
    # bypassing Python sanity checks. We need to handle this case properly.
    # The proper way to handle that is to allow to register a callback function to run in the subprocess at their
    # startup, and warn the caller that this is the master process and that (probably) nothing should be done.
    if uwsgi.numproc > 1 and not uwsgi.opt.get("lazy-apps") and uwsgi.worker_id() == 0:

        if not uwsgi.opt.get("master"):
            # Having multiple workers without the master process is not supported:
            # the postfork hooks are not available, so there's no way to start a different profiler in each
            # worker
            raise uWSGIConfigError("master option must be enabled when multiple processes are used")

        # Register the function to be called in child process at startup
        if worker_callback is not None:
            try:
                import uwsgidecorators
            except ImportError:
                raise uWSGIConfigError("Running under uwsgi but uwsgidecorators cannot be imported")
            uwsgidecorators.postfork(worker_callback)

        if atexit is not None:

            original_atexit = getattr(uwsgi, "atexit", None)

            def _atexit():
                try:
                    atexit()
                except Exception:
                    pass

                if original_atexit is not None:
                    original_atexit()

            uwsgi.atexit = _atexit

        raise uWSGIMasterProcess()



import os
import re
import sys
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import Callable
    from typing import Dict
    from typing import Iterable
    from typing import List
    from typing import Tuple
    from typing import Union

    from ddtrace import Span

import graphql
from graphql import MiddlewareManager
from graphql.error import GraphQLError
from graphql.execution import ExecutionResult
from graphql.language.source import Source

from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import ERROR_MSG
from ddtrace.constants import ERROR_TYPE
from ddtrace.constants import SPAN_MEASURED_KEY
from ddtrace.internal.compat import stringify
from ddtrace.internal.utils import ArgumentError
from ddtrace.internal.utils import get_argument_value
from ddtrace.internal.utils import set_argument_value
from ddtrace.internal.utils.formats import asbool
from ddtrace.internal.utils.version import parse_version
from ddtrace.internal.wrapping import unwrap
from ddtrace.internal.wrapping import wrap
from ddtrace.pin import Pin

from .. import trace_utils
from ...ext import SpanTypes


_graphql_version = parse_version(getattr(graphql, "__version__"))

if _graphql_version < (3, 0):
    from graphql.language.ast import Document
else:
    from graphql.language.ast import DocumentNode as Document


config._add(
    "graphql",
    dict(
        _default_service="graphql",
        resolvers_enabled=asbool(os.getenv("DD_TRACE_GRAPHQL_RESOLVERS_ENABLED", default=False)),
    ),
)

_GRAPHQL_SOURCE = "graphql.source"
_GRAPHQL_OPERATION_TYPE = "graphql.operation.type"
_GRAPHQL_OPERATION_NAME = "graphql.operation.name"


def patch():
    if getattr(graphql, "_datadog_patch", False):
        return
    setattr(graphql, "_datadog_patch", True)
    Pin().onto(graphql)

    for module_str, func_name, wrapper in _get_patching_candidates():
        _update_patching(wrap, module_str, func_name, wrapper)


def unpatch():
    if not getattr(graphql, "_datadog_patch", False) or _graphql_version < (2, 0):
        return

    for module_str, func_name, wrapper in _get_patching_candidates():
        _update_patching(unwrap, module_str, func_name, wrapper)

    setattr(graphql, "_datadog_patch", False)


def _get_patching_candidates():
    if _graphql_version < (3, 0):
        return [
            ("graphql.graphql", "execute_graphql", _traced_query),
            ("graphql.language.parser", "parse", _traced_parse),
            ("graphql.validation.validation", "validate", _traced_validate),
            ("graphql.execution.executor", "execute", _traced_execute),
        ]
    return [
        ("graphql.graphql", "graphql_impl", _traced_query),
        ("graphql.language.parser", "parse", _traced_parse),
        ("graphql.validation.validate", "validate", _traced_validate),
        ("graphql.execution.execute", "execute", _traced_execute),
    ]


def _update_patching(operation, module_str, func_name, wrapper):
    module = sys.modules[module_str]
    func = getattr(module, func_name)
    operation(func, wrapper)


def _traced_parse(func, args, kwargs):
    pin = Pin.get_from(graphql)
    if not pin or not pin.enabled():
        return func(*args, **kwargs)

    source = get_argument_value(args, kwargs, 0, "source")
    source_str = _get_source_str(source)
    # If graphql.parse() is called outside graphql.graphql(), graphql.parse will
    # be a top level span. Therefore we must explicitly set the service name.
    with pin.tracer.trace(
        name="graphql.parse",
        service=trace_utils.int_service(pin, config.graphql),
        span_type=SpanTypes.GRAPHQL,
    ) as span:
        span._set_str_tag(_GRAPHQL_SOURCE, source_str)
        return func(*args, **kwargs)


def _traced_validate(func, args, kwargs):
    pin = Pin.get_from(graphql)
    if not pin or not pin.enabled():
        return func(*args, **kwargs)

    document = get_argument_value(args, kwargs, 1, "ast")
    source_str = _get_source_str(document)
    # If graphql.validate() is called outside graphql.graphql(), graphql.validate will
    # be a top level span. Therefore we must explicitly set the service name.
    with pin.tracer.trace(
        name="graphql.validate",
        service=trace_utils.int_service(pin, config.graphql),
        span_type=SpanTypes.GRAPHQL,
    ) as span:
        span._set_str_tag(_GRAPHQL_SOURCE, source_str)
        errors = func(*args, **kwargs)
        _set_span_errors(errors, span)
        return errors


def _traced_execute(func, args, kwargs):
    pin = Pin.get_from(graphql)
    if not pin or not pin.enabled():
        return func(*args, **kwargs)

    if config.graphql.resolvers_enabled:
        # patch resolvers
        args, kwargs = _inject_trace_middleware_to_args(_resolver_middleware, args, kwargs)

    # set resource name
    if _graphql_version < (3, 0):
        document = get_argument_value(args, kwargs, 1, "document_ast")
    else:
        document = get_argument_value(args, kwargs, 1, "document")
    source_str = _get_source_str(document)

    with pin.tracer.trace(
        name="graphql.execute",
        resource=source_str,
        service=trace_utils.int_service(pin, config.graphql),
        span_type=SpanTypes.GRAPHQL,
    ) as span:
        _set_span_operation_tags(span, document)
        span._set_str_tag(_GRAPHQL_SOURCE, source_str)

        result = func(*args, **kwargs)
        if isinstance(result, ExecutionResult):
            # set error tags if the result contains a list of GraphqlErrors, skip if it's a promise
            _set_span_errors(result.errors, span)
        return result


def _traced_query(func, args, kwargs):
    pin = Pin.get_from(graphql)
    if not pin or not pin.enabled():
        return func(*args, **kwargs)

    # set resource name
    source = get_argument_value(args, kwargs, 1, "source")
    resource = _get_source_str(source)

    with pin.tracer.trace(
        name="graphql.request",
        resource=resource,
        service=trace_utils.int_service(pin, config.graphql),
        span_type=SpanTypes.GRAPHQL,
    ) as span:
        # mark span as measured and set sample rate
        span.set_tag(SPAN_MEASURED_KEY)
        sample_rate = config.graphql.get_analytics_sample_rate()
        if sample_rate is not None:
            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)

        result = func(*args, **kwargs)
        if isinstance(result, ExecutionResult):
            # set error tags if the result contains a list of GraphqlErrors, skip if it's a promise
            # If the wrapped validate and execute functions return a list of errors we will duplicate
            # the span errors here.
            _set_span_errors(result.errors, span)
        return result


def _resolver_middleware(next_middleware, root, info, **args):
    """
    trace middleware which wraps the resolvers of graphql fields.
    Note - graphql middlewares can not be a partial. It must be a class or a function.
    """
    pin = Pin.get_from(graphql)
    if not pin or not pin.enabled():
        return next_middleware(root, info, **args)

    with pin.tracer.trace(
        name="graphql.resolve",
        resource=info.field_name,
        span_type=SpanTypes.GRAPHQL,
    ):
        return next_middleware(root, info, **args)


def _inject_trace_middleware_to_args(trace_middleware, args, kwargs):
    # type: (Callable, Tuple, Dict) -> Tuple[Tuple, Dict]
    """
    Adds a trace middleware to graphql.execute(..., middleware, ...)
    """
    middlewares_arg = 8
    if _graphql_version >= (3, 2):
        # middleware is the 10th argument graphql.execute(..) version 3.2+
        middlewares_arg = 9

    # get middlewares from args or kwargs
    try:
        middlewares = get_argument_value(args, kwargs, middlewares_arg, "middleware") or []
        if isinstance(middlewares, MiddlewareManager):
            # First we must get the middlewares iterable from the MiddlewareManager then append
            # trace_middleware. For the trace_middleware to be called a new MiddlewareManager will
            # need to initialized. This is handled in graphql.execute():
            # https://github.com/graphql-python/graphql-core/blob/v3.2.1/src/graphql/execution/execute.py#L254
            middlewares = middlewares.middlewares  # type: Iterable
    except ArgumentError:
        middlewares = []

    # Note - graphql middlewares are called in reverse order
    # add trace_middleware to the end of the list to wrap the execution of resolver and all middlewares
    middlewares = list(middlewares) + [trace_middleware]

    # update args and kwargs to contain trace_middleware
    args, kwargs = set_argument_value(args, kwargs, middlewares_arg, "middleware", middlewares)
    return args, kwargs


def _get_source_str(obj):
    # type: (Union[str, Source, Document]) -> str
    """
    Parses graphql Documents and Source objects to retrieve
    the graphql source input for a request.
    """
    if isinstance(obj, str):
        source_str = obj
    elif isinstance(obj, Source):
        source_str = obj.body
    elif isinstance(obj, Document):
        source_str = obj.loc.source.body
    else:
        source_str = ""
    # remove new lines, tabs and extra whitespace from source_str
    return re.sub(r"\s+", " ", source_str).strip()


def _set_span_errors(errors, span):
    # type: (List[GraphQLError], Span) -> None
    if not errors:
        # do nothing if the list of graphql errors is empty
        return

    span.error = 1
    exc_type_str = "%s.%s" % (GraphQLError.__module__, GraphQLError.__name__)
    span._set_str_tag(ERROR_TYPE, exc_type_str)
    error_msgs = "\n".join([stringify(error) for error in errors])
    # Since we do not support adding and visualizing multiple tracebacks to one span
    # we will not set the error.stack tag on graphql spans. Setting only one traceback
    # could be misleading and might obfuscate errors.
    span._set_str_tag(ERROR_MSG, error_msgs)


def _set_span_operation_tags(span, document):
    operation_def = graphql.get_operation_ast(document)
    if not operation_def:
        return

    # operation_def.operation should never be None
    if _graphql_version < (3, 0):
        span._set_str_tag(_GRAPHQL_OPERATION_TYPE, operation_def.operation)
    else:
        # OperationDefinition.operation is an Enum in graphql-core>=3
        span._set_str_tag(_GRAPHQL_OPERATION_TYPE, operation_def.operation.value)

    if operation_def.name:
        span._set_str_tag(_GRAPHQL_OPERATION_NAME, operation_def.name.value)



import os

import psycopg2
import pytest

from ddtrace.contrib.psycopg.patch import patch
from ddtrace.contrib.psycopg.patch import unpatch
from ddtrace.vendor import wrapt
from tests.contrib.config import POSTGRES_CONFIG
from tests.utils import override_config


@pytest.fixture(autouse=True)
def patch_psycopg():
    patch()
    assert isinstance(psycopg2.connect, wrapt.ObjectProxy)
    yield
    unpatch()


@pytest.mark.snapshot()
def test_connect_default():
    """By default we do not trace psycopg2.connect method"""
    conn = psycopg2.connect(**POSTGRES_CONFIG)
    assert conn


@pytest.mark.snapshot()
def test_connect_traced():
    """When explicitly enabled, we trace psycopg2.connect method"""
    with override_config("psycopg", {"trace_connect": True}):
        conn = psycopg2.connect(**POSTGRES_CONFIG)
        assert conn


@pytest.mark.snapshot(token="tests.contrib.psycopg.test_psycopg_snapshot.test_connect_traced")
def test_connect_traced_via_env(run_python_code_in_subprocess):
    """When explicitly enabled, we trace psycopg2.connect method"""

    code = """
import psycopg2

import ddtrace
from tests.contrib.config import POSTGRES_CONFIG

ddtrace.patch_all()

conn = psycopg2.connect(**POSTGRES_CONFIG)
assert conn
    """

    env = os.environ.copy()
    env["DD_PSYCOPG_TRACE_CONNECT"] = "true"
    out, err, status, pid = run_python_code_in_subprocess(code, env=env)
    assert status == 0, err
    assert out == b"", err
    assert err == b""



from typing import FrozenSet


PROPAGATION_STYLE_DATADOG = "datadog"
PROPAGATION_STYLE_B3 = "b3"
PROPAGATION_STYLE_B3_SINGLE_HEADER = "b3 single header"
PROPAGATION_STYLE_ALL = frozenset(
    [PROPAGATION_STYLE_DATADOG, PROPAGATION_STYLE_B3, PROPAGATION_STYLE_B3_SINGLE_HEADER]
)  # type: FrozenSet[str]


DEFAULT_SERVICE_NAME = "unnamed_python_service"



import aiohttp
import pytest

from ddtrace.contrib.aiohttp.middlewares import trace_app
from ddtrace.internal.utils import version

from .app.web import setup_app


if version.parse_version(aiohttp.__version__) < (3, 0, 0):

    @pytest.fixture
    def aiohttp_client(test_client):
        return test_client

    @pytest.fixture
    def app_tracer(tracer, loop):
        app = setup_app()
        trace_app(app, tracer)
        return app, tracer

    @pytest.fixture
    def patched_app_tracer(app_tracer):
        app, tracer = app_tracer
        return app, tracer
        # When Python 3.5 is dropped, rather do:
        # yield app, tracer
        # unpatch()

    @pytest.fixture
    def untraced_app_tracer(tracer, loop):
        app = setup_app()
        return app, tracer
        # When Python 3.5 is dropped, rather do:
        # yield app, tracer
        # unpatch()


else:

    @pytest.fixture
    async def app_tracer(tracer, loop):
        app = setup_app()
        trace_app(app, tracer)
        return app, tracer

    @pytest.fixture
    async def patched_app_tracer(app_tracer):
        app, tracer = app_tracer
        return app, tracer
        # When Python 3.5 is dropped, rather do:
        # yield app, tracer
        # unpatch()

    @pytest.fixture
    async def untraced_app_tracer(tracer, loop):
        app = setup_app()
        return app, tracer
        # When Python 3.5 is dropped, rather do:
        # yield app, tracer
        # unpatch()



from contextlib import contextmanager
from random import shuffle
import sys

import mock
import pytest
from six import PY2

from ddtrace.internal.injection import InvalidLine
from ddtrace.internal.injection import eject_hook
from ddtrace.internal.injection import eject_hooks
from ddtrace.internal.injection import inject_hook
from ddtrace.internal.injection import inject_hooks
from ddtrace.internal.utils.inspection import linenos


@contextmanager
def injected_hook(f, hook, arg):
    code = f.__code__
    line = min(linenos(f))

    inject_hook(f, hook, line, arg)

    yield f

    eject_hook(f, hook, line, arg)

    if sys.version_info[:2] not in {(3, 5), (3, 6)}:
        assert f.__code__ == code
    assert f.__code__ is not code


def injection_target(a, b):
    a = a ^ b
    b = a ^ b
    # comment
    a = a ^ b

    return a, b


def loop_target(n):
    for i in range(n):
        a, b = injection_target(n, n + 1)


def generator_target(n):
    yield "begin"
    # yield from range(n)
    yield "end"
    return


def multiline(new_target):
    int(
        "0x2a",
        base=16,
    )


def test_inject_hook():
    hook = mock.Mock()

    lo = min(linenos(injection_target))
    result = inject_hook(injection_target, hook, lo, 42)

    assert result(1, 2) == (2, 1)

    hook.assert_called_once_with(42)


def test_eject_hook():
    hook = mock.Mock()

    lo = min(linenos(injection_target))
    result = eject_hook(inject_hook(injection_target, hook, lo, hook), hook, lo, hook)

    assert result(1, 2) == (2, 1)

    hook.assert_not_called()


def test_inject_hooks():
    n = 2
    hooks = [mock.Mock("hook%d" % _) for _ in range(n)]

    lo = min(linenos(injection_target))
    lines = list(range(lo, lo + n))
    idents = list(range(42, 42 + n))

    failed = inject_hooks(injection_target, list(zip(hooks, lines, idents)))
    assert failed == []

    assert injection_target(1, 2) == (2, 1)

    for hook, ident in zip(hooks, idents):
        hook.assert_called_with(ident)


def test_inject_hooks_some_invalid():
    lines = list(linenos(injection_target))
    lines.append(max(lines) + 10)
    n = len(lines)
    hooks = [mock.Mock("hook%d" % _) for _ in range(n)]

    idents = list(range(42, 42 + n))

    failed = inject_hooks(injection_target, list(zip(hooks, lines, idents)))
    assert failed == [(hooks[-1], lines[-1], idents[-1])]

    assert injection_target(1, 2) == (2, 1)

    for hook, ident in zip(hooks[:-1], idents):
        hook.assert_called_with(ident)


def test_eject_hooks():
    hooks = [mock.Mock("hook%d" % _) for _ in range(2)]

    lo = min(linenos(injection_target))
    lines = list(range(lo, lo + 2))
    hook_data = list(zip(hooks, lines, hooks))
    failed = inject_hooks(injection_target, hook_data)
    assert failed == []

    failed = eject_hooks(injection_target, hook_data)
    assert failed == []

    assert injection_target(1, 2) == (2, 1)

    for hook in hooks:
        hook.assert_not_called()


def test_eject_hooks_some_invalid():
    hooks = [mock.Mock("hook%d" % _) for _ in range(2)]

    lo = min(linenos(injection_target))
    lines = list(range(lo, lo + 2))
    hook_data = list(zip(hooks, lines, hooks))
    failed = inject_hooks(injection_target, hook_data)
    assert failed == []

    invalid_hook = (hooks[-1], lines[-1] + 200, hooks[-1])
    hook_data.append(invalid_hook)
    failed = eject_hooks(injection_target, hook_data)
    assert failed == [invalid_hook]

    assert injection_target(1, 2) == (2, 1)

    for hook in hooks:
        hook.assert_not_called()


def test_eject_hooks_same_line():
    hooks = [mock.Mock("hook%d" % _) for _ in range(3)]

    lo = min(linenos(injection_target)) + 1
    lines = [lo] * 3
    shuffled_lines = list(lines)
    shuffle(shuffled_lines)

    failed = inject_hooks(injection_target, list(zip(hooks, lines, hooks)))
    assert failed == []

    failed = eject_hooks(injection_target, list(zip(hooks, shuffled_lines, hooks)))
    assert failed == []

    assert injection_target(1, 2) == (2, 1)

    for hook in hooks:
        hook.assert_not_called()


def test_inject_out_of_bounds():
    with pytest.raises(InvalidLine):
        inject_hook(injection_target, lambda x: x, 0, 0)


def test_inject_invalid_line():
    with pytest.raises(InvalidLine):
        ls = linenos(injection_target)
        gaps = set(range(min(ls), max(ls) + 1)) - ls
        inject_hook(injection_target, lambda x: x, gaps.pop(), 0)


def test_inject_instance_method():
    from tests.submod.stuff import Stuff

    lo = min(linenos(Stuff.instancestuff))
    hook = mock.Mock()
    old_method = Stuff.instancestuff
    if PY2:
        inject_hook(Stuff.instancestuff.__func__, hook, lo, 0)
    else:
        inject_hook(Stuff.instancestuff, hook, lo, 0)

    stuff = Stuff()
    assert stuff.instancestuff(42) == 42
    hook.assert_called_once_with(0)

    Stuff.instancestuff = old_method


def test_inject_in_loop():
    lo = min(linenos(loop_target)) + 1
    hook = mock.Mock()

    n = 10
    new_loop_target = inject_hook(loop_target, hook, lo, 42)
    new_loop_target(n)

    assert hook.call_count == n


def test_inject_in_generator():
    lo = next(iter(linenos(generator_target)))
    hook = mock.Mock()

    new_generator_target = inject_hook(generator_target, hook, lo, 42)
    list(new_generator_target(42))

    assert hook.call_count == 1


def test_inject_in_multiline():
    lo = min(linenos(multiline)) + 2
    hook = mock.Mock()

    new_target = inject_hook(multiline, hook, lo, 42)
    new_target(new_target)

    assert hook.call_count == 1


def test_property():
    stuff = sys.modules["tests.submod.stuff"]

    f = stuff.Stuff.propertystuff.fget

    hook, arg = mock.Mock(), mock.Mock()

    with injected_hook(f, hook, arg):
        stuff.Stuff().propertystuff
    stuff.Stuff().propertystuff

    hook.assert_called_once_with(arg)



from __future__ import absolute_import

import threading
import typing

import attr

from . import _lock
from .. import event


@event.event_class
class ThreadingLockAcquireEvent(_lock.LockAcquireEvent):
    """A threading.Lock has been acquired."""


@event.event_class
class ThreadingLockReleaseEvent(_lock.LockReleaseEvent):
    """A threading.Lock has been released."""


class _ProfiledThreadingLock(_lock._ProfiledLock):

    ACQUIRE_EVENT_CLASS = ThreadingLockAcquireEvent
    RELEASE_EVENT_CLASS = ThreadingLockReleaseEvent


@attr.s
class ThreadingLockCollector(_lock.LockCollector):
    """Record threading.Lock usage."""

    PROFILED_LOCK_CLASS = _ProfiledThreadingLock

    def _get_original(self):
        # type: (...) -> typing.Any
        return threading.Lock

    def _set_original(
        self, value  # type: typing.Any
    ):
        # type: (...) -> None
        threading.Lock = value  # type: ignore[misc]



import graphql
import pytest

from ddtrace import tracer
from ddtrace.contrib.graphql import patch
from ddtrace.contrib.graphql import unpatch
from ddtrace.contrib.graphql.patch import _graphql_version as graphql_version
from tests.utils import override_config
from tests.utils import snapshot


@pytest.fixture(autouse=True)
def enable_graphql_patching():
    patch()
    yield
    unpatch()


@pytest.fixture
def enable_graphql_resolvers():
    with override_config("graphql", dict(resolvers_enabled=True)):
        yield


@pytest.fixture
def test_schema():
    return graphql.GraphQLSchema(
        query=graphql.GraphQLObjectType(
            name="RootQueryType",
            fields={"hello": graphql.GraphQLField(graphql.GraphQLString, None, lambda obj, info: "friend")},
        )
    )


@pytest.fixture
def test_source_str():
    return "query HELLO { hello }"


@pytest.fixture
def test_middleware():
    def dummy_middleware(next_middleware, root, info, **args):
        with tracer.trace("test_middleware"):
            return next_middleware(root, info, **args)

    yield dummy_middleware


@pytest.fixture
def test_source(test_source_str):
    return graphql.Source(test_source_str)


@pytest.mark.asyncio
async def test_graphql(test_schema, test_source_str, snapshot_context):
    with snapshot_context():
        if graphql_version < (3, 0):
            result = graphql.graphql(test_schema, test_source_str)
        else:
            result = await graphql.graphql(test_schema, test_source_str)
        assert result.data == {"hello": "friend"}


@pytest.mark.asyncio
async def test_graphql_with_traced_resolver(test_schema, test_source_str, snapshot_context, enable_graphql_resolvers):
    with snapshot_context():
        if graphql_version < (3, 0):
            result = graphql.graphql(test_schema, test_source_str)
        else:
            result = await graphql.graphql(test_schema, test_source_str)
        assert result.data == {"hello": "friend"}


@pytest.mark.asyncio
async def test_graphql_error(test_schema, snapshot_context):
    with snapshot_context(ignores=["meta.error.type", "meta.error.msg"]):
        if graphql_version < (3, 0):
            result = graphql.graphql(test_schema, "{ invalid_schema }")
        else:
            result = await graphql.graphql(test_schema, "{ invalid_schema }")
        assert len(result.errors) == 1
        assert isinstance(result.errors[0], graphql.error.GraphQLError)
        assert "Cannot query field" in result.errors[0].message


@snapshot(token_override="tests.contrib.graphql.test_graphql.test_graphql")
@pytest.mark.skipif(graphql_version >= (3, 0), reason="graphql version>=3.0 does not return a promise")
def test_graphql_v2_promise(test_schema, test_source_str):
    promise = graphql.graphql(test_schema, test_source_str, return_promise=True)
    result = promise.get()
    assert result.data == {"hello": "friend"}


@snapshot(
    token_override="tests.contrib.graphql.test_graphql.test_graphql_error",
    ignores=["meta.error.type", "meta.error.msg"],
)
@pytest.mark.skipif(graphql_version >= (3, 0), reason="graphql.graphql is NOT async in v2.0")
def test_graphql_error_v2_promise(test_schema):
    promise = graphql.graphql(test_schema, "{ invalid_schema }", return_promise=True)
    result = promise.get()
    assert len(result.errors) == 1
    assert isinstance(result.errors[0], graphql.error.GraphQLError)
    assert result.errors[0].message == 'Cannot query field "invalid_schema" on type "RootQueryType".'


@snapshot()
@pytest.mark.skipif(graphql_version >= (3, 0), reason="graphql.graphql is NOT async in v2.0")
def test_graphql_v2_with_document(test_schema, test_source_str):
    source = graphql.language.source.Source(test_source_str, "GraphQL request")
    document_ast = graphql.language.parser.parse(source)
    result = graphql.graphql(test_schema, document_ast)
    assert result.data == {"hello": "friend"}


@snapshot(token_override="tests.contrib.graphql.test_graphql.test_graphql")
@pytest.mark.skipif(graphql_version < (3, 0), reason="graphql.graphql_sync is NOT suppoerted in v2.0")
def test_graphql_sync(test_schema, test_source_str):
    result = graphql.graphql_sync(test_schema, test_source_str)
    assert result.data == {"hello": "friend"}


@snapshot()
def test_graphql_execute_with_middleware(test_schema, test_source_str, test_middleware, enable_graphql_resolvers):
    with tracer.trace("test-execute-instrumentation"):
        source = graphql.language.source.Source(test_source_str, "GraphQL request")
        ast = graphql.language.parser.parse(source)
        # execute() can be imported from two modules, ensure both are patched
        res1 = graphql.execute(test_schema, ast, middleware=[test_middleware])
        res2 = graphql.execution.execute(test_schema, ast, middleware=[test_middleware])
        assert res1.data == {"hello": "friend"}
        assert res2.data == {"hello": "friend"}


@snapshot(token_override="tests.contrib.graphql.test_graphql.test_graphql_execute_with_middleware")
@pytest.mark.skipif(graphql_version < (3, 1), reason="graphql.execute_sync is not supported in graphql<3.1")
def test_graphql_execute_sync_with_middlware_manager(
    test_schema, test_source_str, test_middleware, enable_graphql_resolvers
):
    with tracer.trace("test-execute-instrumentation"):
        source = graphql.language.source.Source(test_source_str, "GraphQL request")
        ast = graphql.language.parser.parse(source)
        middleware_manager = graphql.MiddlewareManager(test_middleware)
        # execute_sync() can be imported from two modules, ensure both are patched
        res1 = graphql.execute_sync(test_schema, ast, middleware=middleware_manager)
        res2 = graphql.execution.execute_sync(test_schema, ast, middleware=middleware_manager)
        assert res1.data == {"hello": "friend"}
        assert res2.data == {"hello": "friend"}
